"""Module containing residual plot functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.plotting.plot_irf_dispersion_center import _plot_irf_dispersion_center
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import MinorSymLogLocator
from pyglotaran_extras.plotting.utils import add_cycler_if_not_none
from pyglotaran_extras.plotting.utils import shift_time_axis_by_irf_location

if TYPE_CHECKING:
    import xarray as xr
    from cycler import Cycler
    from matplotlib.axis import Axis


@use_plot_config(exclude_from_config=("cycler",))
def plot_residual(
    res: xr.Dataset,
    ax: Axis,
    linlog: bool = False,
    linthresh: float = 1,
    show_data: bool | None = False,
    cycler: Cycler | None = PlotStyle().cycler,
    show_irf_dispersion_center: bool = True,
    irf_location: float | None = None,
) -> None:
    """Plot data or residual on a 2D contour plot.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    ax : Axis
        Axis to plot on.
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    show_data : bool | None
        Whether to show the input data or residual. If set to ``None`` the plot is skipped
        which improves plotting performance for big datasets. Defaults to False.
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    show_irf_dispersion_center : bool
        Whether to show the the IRF dispersion center as overlay on the residual/data plot.
        Defaults to True.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    """
    if show_data is None:
        ax.text(
            0.5,
            0.5,
            "Skipped",
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=24,
        )
        return

    add_cycler_if_not_none(ax, cycler)
    data = (
        res.data
        if show_data
        else res.weighted_residual
        if "weighted_residual" in res
        else res.residual
    )
    data = shift_time_axis_by_irf_location(data, irf_location, _internal_call=True)
    title = (
        "dataset"
        if show_data
        else "weighted residual"
        if "weighted_residual" in res
        else "residual"
    )
    shape = np.array(data.shape)
    # Handle different dimensionality of data
    if min(shape) == 1:
        dims = data.coords.dims
        data.plot.line(x=dims[shape.argmax()], ax=ax)
    elif min(shape) < 5:
        data.plot(x="time", ax=ax)
    else:
        data.plot(x="time", ax=ax, add_colorbar=False)
    if show_irf_dispersion_center is True:
        _plot_irf_dispersion_center(
            res, ax=ax, spectral_axis="y", cycler=cycler, irf_location=irf_location
        )
        ax.set_xlabel("time")
        ax.legend(loc="upper right")
    if linlog:
        ax.set_xscale("symlog", linthresh=linthresh)
        ax.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))
    ax.set_title(title)



"""Module containing functionality to initialize a case study."""

from __future__ import annotations

import inspect
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from os import PathLike


def setup_case_study(
    output_folder_name: str = "pyglotaran_results",
    results_folder_root: None | str | PathLike[str] = None,
) -> tuple[Path, Path]:
    """Quickly get folders for a case study.

    This is an execution environment independent (works in python script files
    and notebooks, independent of where the python runtime was called from)
    way to get the folder the analysis code resides in and also creates the
    ``results_folder`` in case it didn't exist before.

    Parameters
    ----------
    output_folder_name : str
        Name of the base folder for the results. Defaults to "pyglotaran_results".
    results_folder_root : None | str | PathLike[str]
        The folder where the results named ``output_folder_name`` should be saved to.
        Defaults to None, which results in the users Home folder being used.

    Returns
    -------
    tuple[Path, Path]
        results_folder, script_folder:

        results_folder:
            Folder to be used to save results in of the pattern
            (``results_folder_root`` / ``output_folder_name`` / ``analysis_folder.parent``).
        analysis_folder:
            Folder the script or Notebook resides in.
    """
    analysis_folder = get_script_dir(nesting=1)
    print(f"Setting up case study for folder: {analysis_folder}")  # noqa: T201
    if results_folder_root is None:
        results_folder_root = Path.home() / output_folder_name
    else:
        results_folder_root = Path(str(results_folder_root)) / output_folder_name
    script_folder_rel = analysis_folder.relative_to(analysis_folder.parent)
    results_folder = (results_folder_root / script_folder_rel).resolve()
    results_folder.mkdir(parents=True, exist_ok=True)
    print(f"Results will be saved in: {results_folder}")  # noqa: T201
    return results_folder, analysis_folder.resolve()


def get_script_dir(*, nesting: int = 0) -> Path:
    """Get parent folder a script/Notebook is executed in.

    This is a helper function for cross compatibility with jupyter notebooks.
    In notebooks the global ``__file__`` variable isn't set, thus we need different
    means to get the folder a script is defined in, which doesn't change with the
    current working director the ``python interpreter`` was called from.

    Parameters
    ----------
    nesting : int
        Number to go up in the call stack to get to the initially calling function.
        This is only needed for library code and not for user code. Defaults to 0 (direct call).

    Returns
    -------
    Path
        Path to the folder the script is located in.

    See Also
    --------
    setup_case_study
    """
    calling_frame = inspect.stack()[nesting + 1].frame
    file_var = calling_frame.f_globals.get("__file__", ".")
    file_path = Path(file_var).resolve()
    return file_path if file_var == "." else file_path.parent



"""Module containing type definitions."""

from __future__ import annotations

from collections.abc import Mapping
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import ParamSpec
from typing import TypeAlias
from typing import TypedDict
from typing import TypeVar

import xarray as xr
from glotaran.project.result import Result
from pydantic import BaseModel
from pydantic import RootModel

if TYPE_CHECKING:
    from pyglotaran_extras.plotting.style import ColorCode


class UnsetType:
    """Type for the ``Unset`` singleton."""

    def __repr__(self) -> str:  # noqa: DOC
        """Representation of instances in editors."""
        return "Unset"


Unset = UnsetType()
"""Value to use as default for an arguments where None is a meaningful value.

This way we can prevent regressions.
"""


class CyclerColor(TypedDict):
    """Color value returned by a cycler."""

    color: str | ColorCode


DatasetConvertible: TypeAlias = xr.Dataset | xr.DataArray | str | Path
"""Types of data which can be converted to a dataset."""
ResultLike: TypeAlias = (
    Result | DatasetConvertible | Mapping[str, DatasetConvertible] | Sequence[DatasetConvertible]
)
"""Result like data which can be converted to a per dataset mapping."""


BuiltinSubPlotLabelFormatFunctionKey: TypeAlias = Literal[
    "number", "upper_case_letter", "lower_case_letter"
]
"""Key supported by ``BuiltinLabelFormatFunctions``."""

SubPlotLabelCoordStrs: TypeAlias = Literal[
    "figure points",
    "figure pixels",
    "figure fraction",
    "subfigure points",
    "subfigure pixels",
    "subfigure fraction",
    "axes points",
    "axes pixels",
    "axes fraction",
    "data",
    "polar",
]

SubPlotLabelCoord: TypeAlias = (
    SubPlotLabelCoordStrs | tuple[SubPlotLabelCoordStrs, SubPlotLabelCoordStrs]
)


Param = ParamSpec("Param")
RetType = TypeVar("RetType")

SupportsModelDump = TypeVar("SupportsModelDump", bound=(BaseModel | RootModel[Any]))



"""Module containing configuration."""

from __future__ import annotations

import importlib
import json
import sys
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import PrivateAttr
from pydantic import PydanticUserError
from pydantic import create_model
from pydantic.fields import FieldInfo
from ruamel.yaml import YAML

from pyglotaran_extras.config.plot_config import PlotConfig
from pyglotaran_extras.config.plot_config import PlotLabelOverrideMap
from pyglotaran_extras.config.plot_config import __PlotFunctionRegistry
from pyglotaran_extras.config.utils import add_yaml_repr
from pyglotaran_extras.io.setup_case_study import get_script_dir

if TYPE_CHECKING:
    from collections.abc import Generator
    from collections.abc import Iterable

# Only imported for builtin schema generation
from collections.abc import Sequence  # noqa: F401
from typing import Literal  # noqa: F401

CONFIG_FILE_STEM = "pygta_config"

EXPORT_TEMPLATE = """\
# yaml-language-server: $schema={schema_path}

{config_yaml}\
"""


class UsePlotConfigError(Exception):
    """Error thrown when ``use_plot_config`` has none json serializable kwargs."""

    def __init__(self, func_name: str, error: PydanticUserError) -> None:  # noqa: DOC
        """Use ``func_name`` and original ``error`` to create error message."""
        msg = (
            f"The function ``{func_name}`` decorated with ``use_plot_config`` has an keyword "
            "argument with a type annotation can not be represents in the config.\n"
            "Please use the name of this keyword argument in the ``exclude_from_config`` "
            "keyword argument to ``use_plot_config``.\n"
            f"Original error:\n{error}"
        )
        super().__init__(msg)


@add_yaml_repr
class Config(BaseModel):
    """Main configuration class."""

    model_config = ConfigDict(extra="forbid")

    plotting: PlotConfig = PlotConfig()
    _source_files: list[Path] = PrivateAttr(default_factory=list)
    _source_hash: int = PrivateAttr(default=hash(()))

    def merge(self, other: Config) -> Config:
        """Merge two ``Config``'s where ``other`` overrides values and return a new instance.

        Parameters
        ----------
        other : Config
            Other ``Config`` to merge in.

        Returns
        -------
        Config
        """
        merged = self.model_copy(deep=True)
        merged.plotting = merged.plotting.merge(other.plotting)
        for source_file in other._source_files:
            if source_file in merged._source_files:
                merged._source_files.remove(source_file)
            merged._source_files.append(source_file)
        merged._source_hash = merged._calculate_source_hash()
        return merged

    def _reset(self, other: Config | None = None) -> Config:
        """Reset self to ``other`` config or default initialization.

        Parameters
        ----------
        other : Config | None
            Other ``Config`` to to reset to.

        Returns
        -------
        Config
        """
        if other is None:
            other = Config()
        else:
            self._source_files = other._source_files
        self.plotting = other.plotting
        return self

    def _calculate_source_hash(self) -> int:  # noqa: DOC
        """Calculate hash of source files based on their modification time."""
        return hash(tuple(source_file.stat().st_mtime for source_file in self._source_files))

    def reload(self) -> Config:
        """Reset and reload config from files.

        Returns
        -------
        Config
        """
        if self._source_hash == self._calculate_source_hash():
            return self
        context_config = getattr(self.plotting, "__context_config", None)
        merged = self._reset()
        for config in load_config_files(self._source_files):
            merged = merged.merge(config)
        self.plotting = merged.plotting
        if context_config is not None:
            setattr(self.plotting, "__context_config", context_config)
        self._source_hash = merged._source_hash
        return self

    def load(self, config_file_path: Path | str) -> Config:
        """Disregard current config and config file paths, and reload from ``config_file_path``.

        Parameters
        ----------
        config_file_path : Path | str
            Path to the config file to load.

        Returns
        -------
        Config
        """
        self._source_files = [Path(config_file_path)]
        return self.reload()

    def export(self, export_folder: Path | str | None = None, *, update: bool = True) -> Path:
        """Export current config and schema to ``export_folder``.

        Parameters
        ----------
        export_folder : Path | str | None
            Folder to export config and scheme to. Defaults to None, which means that the script
            folder is used
        update : bool
            Whether to update or overwrite and existing config file. Defaults to True

        Returns
        -------
        Path
            Path to exported config file.
        """
        if export_folder is None:
            from pyglotaran_extras import SCRIPT_DIR

            export_folder = SCRIPT_DIR
        else:
            export_folder = Path(export_folder)
        export_folder.mkdir(parents=True, exist_ok=True)
        schema_path = create_config_schema(export_folder)
        export_path = export_folder / f"{CONFIG_FILE_STEM}.yml"
        if export_path.is_file() is True and update is True:
            merged = Config().load(export_path).merge(self)
            config = merged
        else:
            config = self
        export_path.write_text(
            EXPORT_TEMPLATE.format(schema_path=schema_path.name, config_yaml=config),
            encoding="utf8",
        )
        return export_path

    def rediscover(self, *, include_home_dir: bool = True, lookup_depth: int = 2) -> list[Path]:
        """Rediscover config paths based on the ``SCRIPT_DIR`` discovered on import.

        Parameters
        ----------
        include_home_dir : bool
            Where or not to include the users home folder in the config lookup. Defaults to True
        lookup_depth : int
            Depth at which to look for configs in parent folders of ``script_dir``.
            If set to ``1`` only ``script_dir`` will be considered as config dir.
            Defaults to ``2``.

        Returns
        -------
        list[Path]
            Paths of the discovered config files.
        """
        from pyglotaran_extras import SCRIPT_DIR

        self._source_files = list(
            discover_config_files(
                SCRIPT_DIR, include_home_dir=include_home_dir, lookup_depth=lookup_depth
            )
        )
        return self._source_files

    def init_project(self) -> Config:
        """Initialize configuration for the current project.

        This will use the configs discovered and resolved config during import to create a new
        config and schema for your current project inside of your working directory (script dir),
        if it didn't exist before.

        Returns
        -------
        Config
        """
        from pyglotaran_extras import SCRIPT_DIR

        if any(find_config_in_dir(SCRIPT_DIR)) is False:
            self.export()
        self.rediscover()
        self.reload()
        return self


def find_config_in_dir(dir_path: Path) -> Generator[Path, None, None]:
    """Find the config file inside of dir ``dir_path``.

    Parameters
    ----------
    dir_path : Path
        Directory path to look for a config file.

    Yields
    ------
    Path
    """
    for extension in (".yaml", ".yml"):
        config_file = (dir_path / CONFIG_FILE_STEM).with_suffix(extension)
        if config_file.is_file():
            yield config_file


def discover_config_files(
    script_dir: Path, *, include_home_dir: bool = True, lookup_depth: int = 2
) -> Generator[Path, None, None]:
    """Find config files in the users home folder and the current working dir and parents.

    Parameters
    ----------
    script_dir : Path
        Path to the current scripts/notebooks parent folder.
    include_home_dir : bool
        Where or not to include the users home folder in the config lookup. Defaults to True
    lookup_depth : int
        Depth at which to look for configs in parent folders of ``script_dir``.
        If set to ``1`` only ``script_dir`` will be considered as config dir.
        Defaults to ``2``.

    Yields
    ------
    Path
    """
    if include_home_dir is True:
        yield from find_config_in_dir(Path.home())
    parent_dirs = tuple(reversed((script_dir / "dummy").parents))
    if lookup_depth > 0 and lookup_depth <= len(parent_dirs):
        parent_dirs = parent_dirs[-lookup_depth:]
    for parent in parent_dirs:
        yield from find_config_in_dir(parent)


def load_config_files(config_paths: Iterable[Path]) -> Generator[Config, None, None]:
    """Load config files into new config instances.

    Parameters
    ----------
    config_paths : Iterable[Path]
        Path to the config file.

    Yields
    ------
    Config
    """
    yaml = YAML()
    for config_path in config_paths:
        try:
            config_dict = yaml.load(config_path)
            config = Config.model_validate(config_dict) if config_dict is not None else Config()
            config._source_files.append(config_path)
            yield config
        # We use a very broad range of exception to ensure the config loading at import never
        # breaks importing
        except Exception as error:  # noqa: BLE001
            print(  # noqa: T201
                "Error loading the config:\n",
                f"Source path: {config_path.as_posix()}\n",
                f"Error: {error}",
                file=sys.stderr,
                sep="",
            )


def merge_configs(configs: Iterable[Config]) -> Config:
    """Merge ``Config``'s from left to right, where the right ``Config`` overrides the left.

    Parameters
    ----------
    configs : Iterable[Config]
        Config instances to merge together.

    Returns
    -------
    Config
    """
    full_config = Config()
    for config in configs:
        full_config = full_config.merge(config)
    return full_config


def load_config(
    script_dir: Path, *, include_home_dir: bool = True, lookup_depth: int = 2
) -> Config:
    """Discover and load config files.

    Parameters
    ----------
    script_dir : Path
        Path to the current scripts/notebooks parent folder.
    include_home_dir : bool
        Where or not to include the users home folder in the config lookup. Defaults to True
    lookup_depth : int
        Depth at which to look for configs in parent folders of ``script_dir``.
        If set to ``1`` only ``script_dir`` will be considered as config dir.
        Defaults to ``2``.

    Returns
    -------
    Config

    See Also
    --------
    discover_config_files
    """
    config_paths = discover_config_files(
        script_dir, include_home_dir=include_home_dir, lookup_depth=lookup_depth
    )
    configs = load_config_files(config_paths)
    return merge_configs(configs)


def _find_script_dir_at_import(package_root_file: str) -> Path:
    """Find the script dir when importing ``pyglotaran_extras``.

    The assumption is that the first file not inside of ``pyglotaran_extras`` or importlib
    is the script in question.
    The max ``nesting_offset`` of 20 was chosen semi arbitrarily (typically ``nesting + offset``
    is around 9-13 depending on the import) to ensure that there won't be an infinite loop.

    Parameters
    ----------
    package_root_file : str
        The dunder file attribute (``__file__``) in the package root file.

    Returns
    -------
    Path
    """
    nesting_offset = 0
    importlib_path = Path(importlib.__file__).parent
    package_root = Path(package_root_file).parent
    script_dir = get_script_dir(nesting=2)
    while (
        importlib_path in (script_dir / "dummy").parents
        or package_root in (script_dir / "dummy").parents
    ) and nesting_offset < 20:
        nesting_offset += 1
        script_dir = get_script_dir(nesting=2 + nesting_offset)
    return script_dir


def create_config_schema(
    output_folder: Path | str | None = None,
    file_name: Path | str = f"{CONFIG_FILE_STEM}.schema.json",
) -> Path:
    """Create json schema file to be used for autocompletion and linting of the config.

    Parameters
    ----------
    output_folder : Path | str | None
        Folder to write schema file to. Defaults to None, which means that the script
            folder is used
    file_name : Path | str
        Name of the scheme file. Defaults to "pygta_config.schema.json"

    Returns
    -------
    Path
        Path to the file the schema got saved to.

    Raises
    ------
    UsePlotConfigError
        If any function decorated with ``use_plot_config`` has a keyword argument with a default
        value and a type annotation that can not be serialized into a json schema.
    """
    json_schema = Config.model_json_schema()
    general_kwargs: dict[str, Any] = {}

    for function_name, default_kwargs in __PlotFunctionRegistry.items():
        try:
            name_prefix = "".join([parts.capitalize() for parts in function_name.split("_")])
            fields: Any = {
                kwarg_name: (
                    kwarg_value["annotation"],
                    FieldInfo(
                        default=kwarg_value["default"], description=kwarg_value["docstring"]
                    ),
                )
                for kwarg_name, kwarg_value in default_kwargs.items()
            }
            kwargs_model_name = f"{name_prefix}Kwargs"
            func_kwargs = create_model(
                kwargs_model_name,
                __config__=ConfigDict(extra="forbid"),
                __doc__=(
                    f"Default arguments to use for ``{function_name}``, "
                    "if not specified in function call."
                ),
                **fields,
            )
            config_model_name = f"{name_prefix}Config"
            func_config = create_model(
                config_model_name,
                __config__=ConfigDict(extra="forbid"),
                __doc__=(
                    f"Plot function configuration specific to ``{function_name}`` "
                    "(overrides values in general)."
                ),
                default_args_override=(func_kwargs, {}),
                axis_label_override=(PlotLabelOverrideMap, PlotLabelOverrideMap()),
            )
            func_json_schema = func_config.model_json_schema()
            general_kwargs |= func_json_schema["$defs"][kwargs_model_name]["properties"]
            json_schema["$defs"] |= func_json_schema.pop("$defs")
            json_schema["$defs"][config_model_name] = func_json_schema
            json_schema["$defs"]["PlotConfig"]["properties"][function_name] = {
                "allOf": [{"$ref": f"#/$defs/{config_model_name}"}]
            }
        except PydanticUserError as error:
            raise UsePlotConfigError(function_name, error)  # noqa: B904
    json_schema["$defs"]["PerFunctionPlotConfig"]["properties"]["default_args_override"][
        "properties"
    ] = general_kwargs
    json_schema["$defs"]["PerFunctionPlotConfig"]["properties"]["default_args_override"][
        "additionalProperties"
    ] = False
    if output_folder is None:
        from pyglotaran_extras import SCRIPT_DIR

        output_folder = SCRIPT_DIR
    else:
        output_folder = Path(output_folder)
    output_folder.mkdir(parents=True, exist_ok=True)
    output_file = output_folder / file_name
    output_file.write_text(json.dumps(json_schema, ensure_ascii=False), encoding="utf8")
    return output_file



"""Data loading utility module."""

from __future__ import annotations

from pathlib import Path
from warnings import warn

import xarray as xr
from glotaran.io import load_dataset
from glotaran.project.result import Result

from pyglotaran_extras.types import DatasetConvertible


def load_data(
    result: DatasetConvertible | Result, dataset_name: str | None = None, *, _stacklevel: int = 2
) -> xr.Dataset:
    """Extract a single dataset from a :class:`DatasetConvertible` object.

    Parameters
    ----------
    result : DatasetConvertible | Result
        Result class instance, xarray Dataset or path to a dataset file.
    dataset_name : str | None
        Name of a specific dataset contained in ``result``, if not provided
        the first dataset will be extracted. Defaults to None.
    _stacklevel : int
        Stacklevel of the warning which is raised when ``result`` is of class ``Result``,
        contains multiple datasets and no ``dataset_name`` is provided. Changing this value is
        only required if you use this function inside of another function. Defaults to 2

    Returns
    -------
    xr.Dataset
        Extracted dataset.

    Raises
    ------
    TypeError
        If ``result`` isn't a :class:`DatasetConvertible` object.
    """
    if isinstance(result, xr.Dataset):
        return result
    if isinstance(result, xr.DataArray):
        return result.to_dataset(name="data")
    if isinstance(result, Result):
        if dataset_name is not None:
            return result.data[dataset_name]
        keys = list(result.data)
        if len(keys) > 1:
            warn(
                UserWarning(
                    f"Result contains multiple datasets, auto selecting {keys[0]!r}.\n"
                    f"Pass the dataset set you want to plot (e.g. result.data[{keys[0]!r}]) , "
                    f"to deactivate this Warning.\nPossible dataset names are: {keys}"
                ),
                stacklevel=_stacklevel,
            )
        return result.data[keys[0]]
    if isinstance(result, str | Path):
        return load_data(load_dataset(result))
    msg = f"Result needs to be of type {DatasetConvertible!r}, but was {result!r}."
    raise TypeError(msg)



"""Module containing deprecation functionality."""

from __future__ import annotations

from importlib.metadata import distribution
from warnings import warn

FIG_ONLY_WARNING = (
    "The ``figure_only`` argument was deprecated please remove it from the function call.\n"
    "This usage will be an error in version: 0.8.0."
)


class OverdueDeprecationError(Exception):
    """Error thrown when a deprecation should have been removed.

    See Also
    --------
    deprecate
    warn_deprecated
    deprecate_module_attribute
    deprecate_submodule
    deprecate_dict_entry
    """


class PyglotaranExtrasApiDeprecationWarning(UserWarning):
    """Warning to give users about API changes.

    See Also
    --------
    warn_deprecated
    """


def pyglotaran_extras_version() -> str:
    """Version of the distribution.

    This is basically the same as ``pyglotaran_extras.__version__`` but independent
    from pyglotaran_extras.
    This way all of the deprecation functionality can be used even in
    ``pyglotaran_extras.__init__.py`` without moving the import below the definition of
    ``__version__`` or causing a circular import issue.

    Returns
    -------
    str
        The version string.
    """
    return distribution("pyglotaran-extras").version


def parse_version(version_str: str) -> tuple[int, int, int]:
    """Parse version string to tuple of three ints for comparison.

    Parameters
    ----------
    version_str : str
        Fully qualified version string of the form 'major.minor.patch'.

    Returns
    -------
    tuple[int, int, int]
        Version as tuple.

    Raises
    ------
    ValueError
        If ``version_str`` has less that three elements separated by ``.``.
    ValueError
        If ``version_str`` 's first three elements can not be casted to int.
    """
    error_message = (
        "version_str needs to be a fully qualified version consisting of "
        f"int parts (e.g. '0.0.1'), got {version_str!r}"
    )
    split_version = version_str.partition("-")[0].split(".")
    if len(split_version) < 3:
        raise ValueError(error_message)
    try:
        return tuple(map(int, (*split_version[:2], split_version[2].partition("rc")[0])))  # type:ignore[return-value]
    except ValueError as error:
        raise ValueError(error_message) from error


def check_overdue(deprecated_qual_name_usage: str, to_be_removed_in_version: str) -> None:
    """Check if a deprecation is overdue for removal.

    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.

    Raises
    ------
    OverdueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.
    """
    if (
        parse_version(pyglotaran_extras_version()) >= parse_version(to_be_removed_in_version)
        and "dev" not in pyglotaran_extras_version()
    ):
        msg = (
            f"Support for {deprecated_qual_name_usage.partition('(')[0]!r} "
            f"was supposed to be dropped in version: {to_be_removed_in_version!r}.\n"
            f"Current version is: {pyglotaran_extras_version()!r}"
        )
        raise OverdueDeprecationError(msg)


def warn_deprecated(
    *,
    deprecated_qual_name_usage: str,
    new_qual_name_usage: str,
    to_be_removed_in_version: str,
    stacklevel: int = 2,
) -> None:
    """Raise deprecation warning with change information.

    The change information are old / new usage information and end of support version.

    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    new_qual_name_usage : str
        New usage as fully qualified name e.g.:
        ``'glotaran.io.load_model(model_yml_str, format_name="yml_str")'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.

    stacklevel : int
        Stack at which the warning should be shown as raise. Default: 2

    Raises
    ------
    OverdueDeprecation
        If the current version is greater or equal to ``to_be_removed_in_version``.
    """
    check_overdue(deprecated_qual_name_usage, to_be_removed_in_version)
    warn(
        PyglotaranExtrasApiDeprecationWarning(
            f"Usage of {deprecated_qual_name_usage!r} was deprecated, "
            f"use {new_qual_name_usage!r} instead.\n"
            f"This usage will be an error in version: {to_be_removed_in_version!r}."
        ),
        stacklevel=stacklevel,
    )



"""Module containing a-matrix render functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from glotaran.utils.ipython import MarkdownStr
from tabulate import tabulate

from pyglotaran_extras.inspect.utils import pretty_format_numerical
from pyglotaran_extras.inspect.utils import pretty_format_numerical_iterable
from pyglotaran_extras.inspect.utils import wrap_in_details_tag
from pyglotaran_extras.io.utils import result_dataset_mapping

if TYPE_CHECKING:
    import xarray as xr

    from pyglotaran_extras.types import ResultLike


def a_matrix_to_html_table(
    a_matrix: xr.DataArray,
    megacomplex_suffix: str,
    *,
    normalize_initial_concentration: bool = False,
    decimal_places: int = 3,
) -> str:
    """Create HTML multi header table from a-matrix.

    Parameters
    ----------
    a_matrix : xr.DataArray
        DataArray containing the a-matrix values and coordinates.
    megacomplex_suffix : str
        Megacomplex suffix used for the a-matrix data variable and coordinate names.
    normalize_initial_concentration : bool
        Whether or not to normalize the initial concentration. Defaults to False.
    decimal_places : int
        Decimal places to display. Defaults to 3.

    Returns
    -------
    str
        Multi header HTML table representing the a-matrix.
    """
    species = a_matrix.coords[f"species_{megacomplex_suffix}"].to_numpy()
    # Crete a copy so normalization does not mutate the original values
    initial_concentration = np.array(
        a_matrix.coords[f"initial_concentration_{megacomplex_suffix}"].to_numpy()
    )
    lifetime = a_matrix.coords[f"lifetime_{megacomplex_suffix}"].to_numpy()

    if normalize_initial_concentration is True:
        initial_concentration /= initial_concentration.sum()

    header = (
        ["species<br>initial concentration<br>lifetime↓"]
        + [
            f"{sp}<br>{pretty_format_numerical(ic,decimal_places)}<br>&nbsp;"
            for sp, ic in zip(species, initial_concentration, strict=True)
        ]
        + ["Sum"]
    )

    data = [
        pretty_format_numerical_iterable(
            (lifetime, *amps, amps.sum()), decimal_places=decimal_places
        )
        for lifetime, amps in zip(lifetime, a_matrix.values, strict=True)
    ]
    data.append(
        pretty_format_numerical_iterable(
            ("Sum", *a_matrix.to_numpy().sum(axis=0), a_matrix.to_numpy().sum()),
            decimal_places=decimal_places,
        )
    )

    return (
        tabulate(
            data, headers=header, showindex=False, tablefmt="unsafehtml", disable_numparse=True
        )
        .replace(" 0 ", "   ")
        .replace(" 0<", "  <")
        .replace(">0 ", ">  ")
    )


def show_a_matrixes(
    result: ResultLike,
    *,
    normalize_initial_concentration: bool = False,
    decimal_places: int = 3,
    a_matrix_min_size: int | None = None,
    expanded_datasets: tuple[str, ...] = (),
    heading_offset: int = 2,
) -> MarkdownStr:
    """Show all a-matrixes of a result grouped by dataset and megacomplex name.

    Each dataset is wrapped in a HTML details tag which is by default collapsed.

    Parameters
    ----------
    result : ResultLike
        Result or result dataset.
    normalize_initial_concentration : bool
        Whether or not to normalize the initial concentration. Defaults to False.
    decimal_places : int
        Decimal places to display. Defaults to 3.
    a_matrix_min_size : int | None
        Defaults to None.
    expanded_datasets : tuple[str, ...]
        Names of dataset to expand the details view for. Defaults to empty tuple () which means no
        dataset is expanded.
    heading_offset : int
        Number of heading level to offset the headings. Defaults to 2 which means that the
        first/top most heading is h3.

    Returns
    -------
    MarkdownStr
        Markdown representation of the a-matrixes used in the optimization.
    """
    heading_prefix = heading_offset * "#"
    output_str = f"#{heading_prefix} A-Matrixes\n"

    result_map = result_dataset_mapping(result)

    for dataset_name in result_map:
        a_matrix_names = list(
            filter(
                lambda var_name: var_name.startswith("a_matrix_"),
                result_map[dataset_name].data_vars,
            )
        )

        if not a_matrix_names:
            continue

        details_content = ""
        header_newline_prefix = ""

        for a_matrix_name in a_matrix_names:
            mc_suffix = a_matrix_name.replace("a_matrix_", "")

            a_matrix = result_map[dataset_name][a_matrix_name]

            if a_matrix_min_size is not None and max(a_matrix.shape) < a_matrix_min_size:
                continue

            details_content += f"{header_newline_prefix}###{heading_prefix} {mc_suffix}:\n\n"

            details_content += a_matrix_to_html_table(
                a_matrix,
                mc_suffix,
                normalize_initial_concentration=normalize_initial_concentration,
                decimal_places=decimal_places,
            )
            header_newline_prefix = "\n\n"

        if details_content != "":
            output_str += wrap_in_details_tag(
                details_content,
                summary_content=dataset_name,
                summary_heading_level=2 + heading_offset,
                is_open=dataset_name in expanded_datasets,
            )

    return MarkdownStr(output_str)



"""Pyglotaran extension package with convenience functionality such as plotting."""

from __future__ import annotations

from pyglotaran_extras.config.config import _find_script_dir_at_import
from pyglotaran_extras.config.config import create_config_schema
from pyglotaran_extras.config.config import load_config
from pyglotaran_extras.config.plot_config import PerFunctionPlotConfig
from pyglotaran_extras.config.plot_config import plot_config_context
from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.io.load_data import load_data
from pyglotaran_extras.io.setup_case_study import setup_case_study
from pyglotaran_extras.plotting.plot_coherent_artifact import plot_coherent_artifact
from pyglotaran_extras.plotting.plot_concentrations import plot_concentrations
from pyglotaran_extras.plotting.plot_data import plot_data_overview
from pyglotaran_extras.plotting.plot_doas import plot_doas
from pyglotaran_extras.plotting.plot_guidance import plot_guidance
from pyglotaran_extras.plotting.plot_irf_dispersion_center import plot_irf_dispersion_center
from pyglotaran_extras.plotting.plot_overview import plot_overview
from pyglotaran_extras.plotting.plot_overview import plot_simple_overview
from pyglotaran_extras.plotting.plot_residual import plot_residual
from pyglotaran_extras.plotting.plot_spectra import plot_das
from pyglotaran_extras.plotting.plot_spectra import plot_norm_das
from pyglotaran_extras.plotting.plot_spectra import plot_norm_sas
from pyglotaran_extras.plotting.plot_spectra import plot_sas
from pyglotaran_extras.plotting.plot_spectra import plot_spectra
from pyglotaran_extras.plotting.plot_svd import plot_lsv_data
from pyglotaran_extras.plotting.plot_svd import plot_lsv_residual
from pyglotaran_extras.plotting.plot_svd import plot_rsv_data
from pyglotaran_extras.plotting.plot_svd import plot_rsv_residual
from pyglotaran_extras.plotting.plot_svd import plot_sv_data
from pyglotaran_extras.plotting.plot_svd import plot_sv_residual
from pyglotaran_extras.plotting.plot_svd import plot_svd
from pyglotaran_extras.plotting.plot_traces import plot_fitted_traces
from pyglotaran_extras.plotting.plot_traces import select_plot_wavelengths
from pyglotaran_extras.plotting.utils import add_subplot_labels

__all__ = [
    "load_data",
    "setup_case_study",
    "plot_coherent_artifact",
    "plot_concentrations",
    "plot_data_overview",
    "plot_doas",
    "plot_guidance",
    "plot_irf_dispersion_center",
    "plot_overview",
    "plot_simple_overview",
    "plot_residual",
    "plot_das",
    "plot_norm_das",
    "plot_norm_sas",
    "plot_sas",
    "plot_spectra",
    "plot_lsv_data",
    "plot_lsv_residual",
    "plot_rsv_data",
    "plot_rsv_residual",
    "plot_sv_data",
    "plot_sv_residual",
    "plot_svd",
    "plot_fitted_traces",
    "select_plot_wavelengths",
    "add_subplot_labels",
    # Config
    "PerFunctionPlotConfig",
    "plot_config_context",
    "use_plot_config",
    "create_config_schema",
    "CONFIG",
]

__version__ = "0.7.3"

SCRIPT_DIR = _find_script_dir_at_import(__file__)
"""User script dir determined during import."""
CONFIG = load_config(SCRIPT_DIR)
"""Global config instance."""



"""Module containing SVD plotting functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.io.prepare_dataset import add_svd_to_dataset

from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.deprecation import warn_deprecated
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import MinorSymLogLocator
from pyglotaran_extras.plotting.utils import add_cycler_if_not_none
from pyglotaran_extras.plotting.utils import shift_time_axis_by_irf_location
from pyglotaran_extras.types import Unset
from pyglotaran_extras.types import UnsetType

if TYPE_CHECKING:
    from collections.abc import Sequence

    import xarray as xr
    from cycler import Cycler
    from matplotlib.axis import Axis
    from matplotlib.pyplot import Axes


@use_plot_config(exclude_from_config=("cycler",))
def plot_svd(
    res: xr.Dataset,
    axes: Axes,
    linlog: bool = False,
    linthresh: float = 1,
    cycler: Cycler | None = PlotStyle().cycler,
    nr_of_data_svd_vectors: int = 4,
    nr_of_residual_svd_vectors: int = 2,
    show_data_svd_legend: bool = True,
    show_residual_svd_legend: bool = True,
    irf_location: float | None = None,
    use_svd_number: bool = False,
) -> None:
    """Plot SVD (Singular Value Decomposition) of data and residual.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    axes : Axes
        Axes to plot the SVDs on (needs to be at least 2x3).
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    nr_of_data_svd_vectors : int
        Number of data SVD vector to plot. Defaults to 4.
    nr_of_residual_svd_vectors : int
        Number of residual SVD vector to plot. Defaults to 2.
    show_data_svd_legend : bool
        Whether or not to show the data SVD legend. Defaults to True.
    show_residual_svd_legend : bool
        Whether or not to show the residual SVD legend. Defaults to True.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.
    """
    if "weighted_residual" in res:
        add_svd_to_dataset(dataset=res, name="weighted_residual")
    else:
        add_svd_to_dataset(dataset=res, name="residual")
    plot_lsv_residual(
        res,
        axes[0, 0],
        linlog=linlog,
        linthresh=linthresh,
        cycler=cycler,
        indices=range(nr_of_residual_svd_vectors),
        show_legend=show_residual_svd_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    plot_rsv_residual(
        res,
        axes[0, 1],
        cycler=cycler,
        indices=range(nr_of_residual_svd_vectors),
        show_legend=show_residual_svd_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    plot_sv_residual(res, axes[0, 2], use_svd_number=use_svd_number)
    add_svd_to_dataset(dataset=res, name="data")
    plot_lsv_data(
        res,
        axes[1, 0],
        linlog=linlog,
        linthresh=linthresh,
        cycler=cycler,
        indices=range(nr_of_data_svd_vectors),
        show_legend=show_data_svd_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    plot_rsv_data(
        res,
        axes[1, 1],
        cycler=cycler,
        indices=range(nr_of_data_svd_vectors),
        show_legend=show_data_svd_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    plot_sv_data(res, axes[1, 2], use_svd_number=use_svd_number)


@use_plot_config(exclude_from_config=("cycler",))
def plot_lsv_data(
    res: xr.Dataset,
    ax: Axis,
    indices: Sequence[int] = tuple(range(4)),
    linlog: bool = False,
    linthresh: float = 1,
    cycler: Cycler | None = PlotStyle().cycler,
    show_legend: bool = True,
    irf_location: float | None = None,
    use_svd_number: bool = False,
) -> None:
    """Plot left singular vectors (time) of the data matrix.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    ax : Axis
        Axis to plot on.
    indices : Sequence[int]
        Indices of the singular vector to plot. Defaults to tuple(range(4)).
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    show_legend : bool
        Whether or not to show the legend. Defaults to True.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.
    """
    add_cycler_if_not_none(ax, cycler)
    dLSV = res.data_left_singular_vectors  # noqa: N806
    _plot_svd_vectors(
        vector_data=dLSV,
        indices=indices,
        sv_index_dim="left_singular_value_index",
        ax=ax,
        show_legend=show_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    ax.set_title("data. LSV")
    if linlog:
        ax.set_xscale("symlog", linthresh=linthresh)
        ax.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))


@use_plot_config(exclude_from_config=("cycler",))
def plot_rsv_data(
    res: xr.Dataset,
    ax: Axis,
    indices: Sequence[int] = tuple(range(4)),
    cycler: Cycler | None = PlotStyle().cycler,
    show_legend: bool = True,
    irf_location: float | None = None,
    use_svd_number: bool = False,
) -> None:
    """Plot right singular vectors (spectra) of the data matrix.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    ax : Axis
        Axis to plot on.
    indices : Sequence[int]
        Indices of the singular vector to plot. Defaults to tuple(range(4)).
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    show_legend : bool
        Whether or not to show the legend. Defaults to True.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.
    """
    add_cycler_if_not_none(ax, cycler)
    dRSV = res.data_right_singular_vectors  # noqa: N806
    _plot_svd_vectors(
        vector_data=dRSV,
        indices=indices,
        sv_index_dim="right_singular_value_index",
        ax=ax,
        show_legend=show_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    ax.set_title("data. RSV")


@use_plot_config(exclude_from_config=("cycler",))
def plot_sv_data(
    res: xr.Dataset,
    ax: Axis,
    indices: Sequence[int] = tuple(range(10)),
    cycler: Cycler | None | UnsetType = Unset,
    use_svd_number: bool = False,
) -> None:
    """Plot singular values of the data matrix.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    ax : Axis
        Axis to plot on.
    indices : Sequence[int]
        Indices of the singular vector to plot. Defaults to tuple(range(10)).
    cycler : Cycler | None | UnsetType
        Deprecated since it has no effect. Defaults to Unset.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.
    """
    if cycler is not Unset:
        warn_deprecated(
            deprecated_qual_name_usage="'cycler' argument in 'plot_sv_data'",
            new_qual_name_usage="matplotlib on the axis directly",
            to_be_removed_in_version="0.9.0",
        )
    dSV = res.data_singular_values  # noqa: N806
    x_dim = "singular_value_index"
    if use_svd_number is True:
        x_dim = "singular value number"
        dSV = dSV.assign_coords(  # noqa: N806
            {x_dim: ("singular_value_index", (dSV.singular_value_index + 1).data)}
        )
    dSV.sel({"singular_value_index": list(indices[: len(dSV.singular_value_index)])}).plot.line(
        "ro-", yscale="log", ax=ax, x=x_dim
    )
    ax.set_title("data. log(SV)")


@use_plot_config(exclude_from_config=("cycler",))
def plot_lsv_residual(
    res: xr.Dataset,
    ax: Axis,
    indices: Sequence[int] = tuple(range(2)),
    linlog: bool = False,
    linthresh: float = 1,
    cycler: Cycler | None = PlotStyle().cycler,
    show_legend: bool = True,
    irf_location: float | None = None,
    use_svd_number: bool = False,
) -> None:
    """Plot left singular vectors (time) of the residual matrix.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    ax : Axis
        Axis to plot on.
    indices : Sequence[int]
        Indices of the singular vector to plot. Defaults to tuple(range(4)).
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    show_legend : bool
        Whether or not to show the legend. Defaults to True.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.
    """
    add_cycler_if_not_none(ax, cycler)
    rLSV: xr.DataArray = (  # noqa: N806
        res.weighted_residual_left_singular_vectors
        if "weighted_residual_left_singular_vectors" in res
        else res.residual_left_singular_vectors
    )

    _plot_svd_vectors(
        vector_data=rLSV,
        indices=indices,
        sv_index_dim="left_singular_value_index",
        ax=ax,
        show_legend=show_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    ax.set_title("res. LSV")
    if linlog:
        ax.set_xscale("symlog", linthresh=linthresh)
        ax.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))


@use_plot_config(exclude_from_config=("cycler",))
def plot_rsv_residual(
    res: xr.Dataset,
    ax: Axis,
    indices: Sequence[int] = tuple(range(2)),
    cycler: Cycler | None = PlotStyle().cycler,
    show_legend: bool = True,
    irf_location: float | None = None,
    use_svd_number: bool = False,
) -> None:
    """Plot right singular vectors (spectra) of the residual matrix.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    ax : Axis
        Axis to plot on.
    indices : Sequence[int]
        Indices of the singular vector to plot. Defaults to tuple(range(4)).
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    show_legend : bool
        Whether or not to show the legend. Defaults to True.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.
    """
    add_cycler_if_not_none(ax, cycler)
    rRSV: xr.DataArray = (  # noqa: N806
        res.weighted_residual_right_singular_vectors
        if "weighted_residual_right_singular_vectors" in res
        else res.residual_right_singular_vectors
    )
    _plot_svd_vectors(
        vector_data=rRSV,
        indices=indices,
        sv_index_dim="right_singular_value_index",
        ax=ax,
        show_legend=show_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    ax.set_title("res. RSV")


@use_plot_config(exclude_from_config=("cycler",))
def plot_sv_residual(
    res: xr.Dataset,
    ax: Axis,
    indices: Sequence[int] = tuple(range(10)),
    cycler: Cycler | None | UnsetType = Unset,
    use_svd_number: bool = False,
) -> None:
    """Plot singular values of the residual matrix.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset
    ax : Axis
        Axis to plot on.
    indices : Sequence[int]
        Indices of the singular vector to plot. Defaults to tuple(range(10)).
    cycler : Cycler | None | UnsetType
        Deprecated since it has no effect. Defaults to Unset.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.
    """
    if cycler is not Unset:
        warn_deprecated(
            deprecated_qual_name_usage="'cycler' argument in 'plot_sv_residual'",
            new_qual_name_usage="matplotlib on the axis directly",
            to_be_removed_in_version="0.9.0",
        )
    rSV: xr.DataArray = (  # noqa: N806
        res.weighted_residual_singular_values
        if "weighted_residual_singular_values" in res
        else res.residual_singular_values
    )
    x_dim = "singular_value_index"
    if use_svd_number is True:
        x_dim = "singular value number"
        rSV = rSV.assign_coords(  # noqa: N806
            {x_dim: ("singular_value_index", (rSV.singular_value_index + 1).data)}
        )
    rSV.sel({"singular_value_index": list(indices[: len(rSV.singular_value_index)])}).plot.line(
        "ro-", yscale="log", ax=ax, x=x_dim
    )
    ax.set_title("res. log(SV)")


def _plot_svd_vectors(
    vector_data: xr.DataArray,
    indices: Sequence[int],
    sv_index_dim: str,
    ax: Axis,
    show_legend: bool,
    irf_location: float | None,
    use_svd_number: bool = False,
) -> None:
    """Plot SVD vectors with decreasing zorder on axis ``ax``.

    Parameters
    ----------
    vector_data : xr.DataArray
        DataArray containing the SVD vector data.
    indices : Sequence[int]
        Indices of the singular vector to plot.
    sv_index_dim : str
        Name of the singular value index dimension.
    ax : Axis
        Axis to plot on.
    show_legend : bool
        Whether or not to show the legend.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.

    See Also
    --------
    plot_lsv_data
    plot_rsv_data
    plot_lsv_residual
    plot_rsv_residual
    """
    max_index = len(getattr(vector_data, sv_index_dim))
    values = shift_time_axis_by_irf_location(
        vector_data.isel({sv_index_dim: list(indices[:max_index])}),
        irf_location,
        _internal_call=True,
    )
    x_dim = vector_data.dims[1]
    if x_dim == sv_index_dim:
        values = values.T
        x_dim = vector_data.dims[0]
    for zorder, label, value in zip(range(100)[::-1], indices[:max_index], values, strict=False):
        value.plot.line(
            x=x_dim, ax=ax, zorder=zorder, label=label + 1 if use_svd_number else label
        )
    if show_legend is True:
        ax.legend(title="singular value number" if use_svd_number else sv_index_dim)



"""Module containing functionality to plot fitted traces."""

from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

import matplotlib.pyplot as plt

from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.io.utils import result_dataset_mapping
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import MinorSymLogLocator
from pyglotaran_extras.plotting.utils import PlotDuplicationWarning
from pyglotaran_extras.plotting.utils import add_cycler_if_not_none
from pyglotaran_extras.plotting.utils import add_unique_figure_legend
from pyglotaran_extras.plotting.utils import extract_dataset_scale
from pyglotaran_extras.plotting.utils import extract_irf_location
from pyglotaran_extras.plotting.utils import get_next_cycler_color
from pyglotaran_extras.plotting.utils import select_plot_wavelengths

__all__ = ["select_plot_wavelengths", "plot_fitted_traces"]

if TYPE_CHECKING:
    from collections.abc import Iterable

    from cycler import Cycler
    from matplotlib.axis import Axis
    from matplotlib.figure import Figure
    from matplotlib.pyplot import Axes

    from pyglotaran_extras.types import ResultLike


@use_plot_config(exclude_from_config=("cycler",))
def plot_data_and_fits(
    result: ResultLike,
    wavelength: float,
    axis: Axis,
    center_λ: float | None = None,
    main_irf_nr: int = 0,
    linlog: bool = False,
    linthresh: float = 1,
    divide_by_scale: bool = True,
    per_axis_legend: bool = False,
    y_label: str = "a.u.",
    cycler: Cycler | None = PlotStyle().data_cycler_solid,
    show_zero_line: bool = True,
) -> None:
    """Plot data and fits for a given ``wavelength`` on a given ``axis``.

    If the wavelength isn't part of a dataset, that dataset will be skipped.

    Parameters
    ----------
    result : ResultLike
        Data structure which can be converted to a mapping.
    wavelength : float
        Wavelength to plot data and fits for.
    axis : Axis
        Axis to plot the data and fits on.
    center_λ : float | None
        Center wavelength (λ in nm)
    main_irf_nr : int
        Index of the main ``irf`` component when using an ``irf``
        parametrized with multiple peaks. Defaults to 0.
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    divide_by_scale : bool
        Whether or not to divide the data by the dataset scale used for optimization.
        Defaults to True.
    per_axis_legend : bool
        Whether to use a legend per plot or for the whole figure. Defaults to False.
    y_label : str
        Label used for the y-axis of each subplot.
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().data_cycler_solid.
    show_zero_line : bool
        Whether or not to add a horizontal line at zero. Defaults to True.

    See Also
    --------
    plot_fit_overview
    """
    result_map = result_dataset_mapping(result)
    add_cycler_if_not_none(axis, cycler)
    for dataset_name in result_map:
        if result_map[dataset_name].coords["time"].to_numpy().size == 1:
            continue
        spectral_coords = result_map[dataset_name].coords["spectral"].to_numpy()
        if spectral_coords.min() <= wavelength <= spectral_coords.max():
            result_data = result_map[dataset_name].sel(spectral=[wavelength], method="nearest")
            scale = extract_dataset_scale(result_data, divide_by_scale)
            irf_loc = extract_irf_location(result_data, center_λ, main_irf_nr)
            result_data = result_data.assign_coords(time=result_data.coords["time"] - irf_loc)
            (result_data.data / scale).plot(x="time", ax=axis, label=f"{dataset_name}_data")
            (result_data.fitted_data / scale).plot(x="time", ax=axis, label=f"{dataset_name}_fit")
        else:
            [get_next_cycler_color(axis) for _ in range(2)]
    if linlog:
        axis.set_xscale("symlog", linthresh=linthresh)
        axis.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))
    if show_zero_line is True:
        axis.axhline(0, color="k", linewidth=1)
    axis.set_ylabel(y_label)
    if per_axis_legend is True:
        axis.legend()


@use_plot_config(exclude_from_config=("cycler",))
def plot_fitted_traces(
    result: ResultLike,
    wavelengths: Iterable[float],
    axes_shape: tuple[int, int] = (4, 4),
    center_λ: float | None = None,
    main_irf_nr: int = 0,
    linlog: bool = False,
    linthresh: float = 1,
    divide_by_scale: bool = True,
    per_axis_legend: bool = False,
    figsize: tuple[float, float] = (30, 15),
    title: str = "Fit overview",
    y_label: str = "a.u.",
    cycler: Cycler | None = PlotStyle().data_cycler_solid,
    show_zero_line: bool = True,
) -> tuple[Figure, Axes]:
    """Plot data and their fit in per wavelength plot grid.

    Parameters
    ----------
    result : ResultLike
        Data structure which can be converted to a mapping of datasets.
    wavelengths : Iterable[float]
        Wavelength which should be used for each subplot, should to be of length N*M
        with ``axes_shape`` being of shape (N, M), else it will result in missing plots.
    axes_shape : tuple[int, int]
        Shape of the plot grid (N, M). Defaults to (4, 4).
    center_λ : float | None
        Center wavelength of the IRF (λ in nm).
    main_irf_nr : int
        Index of the main ``irf`` component when using an ``irf``
        parametrized with multiple peaks. Defaults to 0.
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    divide_by_scale : bool
        Whether or not to divide the data by the dataset scale used for optimization.
        Defaults to True.
    per_axis_legend : bool
        Whether to use a legend per plot or for the whole figure. Defaults to False.
    figsize : tuple[float, float]
        Size of the figure (N, M) in inches. Defaults to (30, 15).
    title : str
        Title to add to the figure. Defaults to "Fit overview".
    y_label : str
        Label used for the y-axis of each subplot.
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().data_cycler_solid.
    show_zero_line : bool
        Whether or not to add a horizontal line at zero. Defaults to True.


    Returns
    -------
    tuple[Figure, Axes]
        Figure and axes which can then be refined by the user.

    See Also
    --------
    maximum_coordinate_range
    add_unique_figure_legend
    plot_data_and_fits
    calculate_wavelengths
    """
    result_map = result_dataset_mapping(result)
    fig, axes = plt.subplots(*axes_shape, figsize=figsize)
    nr_of_plots = len(axes.flatten())
    max_spectral_values = max(
        len(result_map[dataset_name].coords["spectral"]) for dataset_name in result_map
    )
    if nr_of_plots > max_spectral_values:
        warn(
            PlotDuplicationWarning(
                f"The number of plots ({nr_of_plots}) exceeds the maximum number of "
                f"spectral data points ({max_spectral_values}), "
                "which will lead in duplicated plots."
            ),
            stacklevel=2,
        )
    for wavelength, axis in zip(wavelengths, axes.flatten(), strict=True):
        plot_data_and_fits(
            result=result_map,
            wavelength=wavelength,
            axis=axis,
            center_λ=center_λ,
            main_irf_nr=main_irf_nr,
            linlog=linlog,
            linthresh=linthresh,
            divide_by_scale=divide_by_scale,
            per_axis_legend=per_axis_legend,
            y_label=y_label,
            cycler=cycler,
            show_zero_line=show_zero_line,
        )
    if per_axis_legend is False:
        add_unique_figure_legend(fig, axes)
    fig.suptitle(title, fontsize=28)
    fig.tight_layout()
    return fig, axes



"""Module containing IRF dispersion plotting functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

import matplotlib.pyplot as plt
import xarray as xr
from matplotlib.axis import Axis
from matplotlib.figure import Figure

from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.io.utils import result_dataset_mapping
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import add_cycler_if_not_none
from pyglotaran_extras.plotting.utils import extract_irf_dispersion_center

if TYPE_CHECKING:
    from typing import Literal

    from cycler import Cycler

    from pyglotaran_extras.types import ResultLike


@use_plot_config(exclude_from_config=("cycler", "ax"))
def plot_irf_dispersion_center(
    result: ResultLike,
    ax: Axis | None = None,
    figsize: tuple[float, float] = (12, 8),
    cycler: Cycler | None = PlotStyle().cycler,
    irf_location: float | None = None,
) -> tuple[Figure, Axis] | None:
    """Plot the IRF dispersion center over the spectral dimension for one or multiple datasets.

    Parameters
    ----------
    result : ResultLike
        Data structure which can be converted to a mapping.
    ax : Axis | None
        Axis to plot on. Defaults to None which means that a new figure and axis will be created.
    figsize : tuple[float, float]
        Size of the figure (N, M) in inches. Defaults to (12, 8).
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.

    Returns
    -------
    tuple[Figure, Axis] | None
        Figure object which contains the plots and the Axis,
        if ``ax`` is not None nothing will be returned.
    """
    result_map = result_dataset_mapping(result)
    if ax is None:
        fig, axis = cast(tuple[Figure, Axis], plt.subplots(1, figsize=figsize))
    else:
        axis = ax
    for dataset_name, dataset in result_map.items():
        _plot_irf_dispersion_center(
            dataset,
            axis,
            spectral_axis="x",
            cycler=cycler,
            label=dataset_name,
            irf_location=irf_location,
        )
    axis.legend()

    if ax is None:
        fig.suptitle("Instrument Response Functions", fontsize=16)
        return fig, axis
    return None


def _plot_irf_dispersion_center(
    res: xr.Dataset,
    ax: Axis,
    *,
    spectral_axis: Literal["x", "y"] = "x",
    cycler: Cycler | None = PlotStyle().cycler,
    label: str = "IRF",
    irf_location: float | None = None,
) -> None:
    """Plot the IRF dispersion center on an Axis ``ax``.

    This is an internal function to be used by higher level functions.

    Parameters
    ----------
    res : xr.Dataset
        Dataset containing the IRF data.
    ax : Axis
        Axis to plot on.
    spectral_axis : Literal["x", "y"]
        Direct of the spectral axis in the plot. Defaults to "x"
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    label : str
        Plot label for the IRF shown in the legend. Defaults to "IRF"
    irf_location : float | None
        Location of the ``irf`` by which the time axis (here values) will get shifted.
        If it is None the time axis will not be shifted. Defaults to None.
    """
    add_cycler_if_not_none(ax, cycler)
    irf = cast(xr.DataArray, extract_irf_dispersion_center(res, as_dataarray=True))
    if irf_location is not None:
        (irf - irf_location).plot(ax=ax, label=label, **{spectral_axis: "spectral"})
    else:
        irf.plot(ax=ax, label=label, **{spectral_axis: "spectral"})



"""Module containing concentration plot functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import MinorSymLogLocator
from pyglotaran_extras.plotting.utils import add_cycler_if_not_none
from pyglotaran_extras.plotting.utils import get_shifted_traces

if TYPE_CHECKING:
    import xarray as xr
    from cycler import Cycler
    from matplotlib.axis import Axis


@use_plot_config(exclude_from_config=("cycler",))
def plot_concentrations(
    res: xr.Dataset,
    ax: Axis,
    center_λ: float | None,
    linlog: bool = False,
    linthresh: float = 1,
    linscale: float = 1,
    main_irf_nr: int = 0,
    cycler: Cycler | None = PlotStyle().cycler,
    title: str = "Concentrations",
) -> None:
    """Plot traces on the given axis ``ax``.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset from a pyglotaran optimization.
    ax : Axis
        Axis to plot the traces on
    center_λ : float | None
        Center wavelength (λ in nm)
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    linscale : float
        This allows the linear range (-linthresh to linthresh) to be stretched
        relative to the logarithmic range.
        Its value is the number of decades to use for each half of the linear range.
        For example, when linscale == 1.0 (the default), the space used for the
        positive and negative halves of the linear range will be equal to one
        decade in the logarithmic range. Defaults to 1.
    main_irf_nr : int
        Index of the main ``irf`` component when using an ``irf``
        parametrized with multiple peaks. Defaults to 0.
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().data_cycler_solid.
    title : str
        Title used for the plot axis. Defaults to "Concentrations".

    See Also
    --------
    get_shifted_traces
    """
    add_cycler_if_not_none(ax, cycler)
    traces = get_shifted_traces(res, center_λ, main_irf_nr)

    if "spectral" in traces.coords:
        traces.sel(spectral=center_λ, method="nearest").plot.line(x="time", ax=ax)
    else:
        traces.plot.line(x="time", ax=ax)
    ax.set_title(title)

    if linlog:
        ax.set_xscale("symlog", linthresh=linthresh, linscale=linscale)
        ax.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))



"""Module containing overview plotting functionality."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from warnings import warn

import matplotlib.pyplot as plt

from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.deprecation.deprecation_utils import FIG_ONLY_WARNING
from pyglotaran_extras.deprecation.deprecation_utils import PyglotaranExtrasApiDeprecationWarning
from pyglotaran_extras.io.load_data import load_data
from pyglotaran_extras.plotting.plot_concentrations import plot_concentrations
from pyglotaran_extras.plotting.plot_guidance import plot_guidance
from pyglotaran_extras.plotting.plot_residual import plot_residual
from pyglotaran_extras.plotting.plot_spectra import plot_sas
from pyglotaran_extras.plotting.plot_spectra import plot_spectra
from pyglotaran_extras.plotting.plot_svd import plot_lsv_residual
from pyglotaran_extras.plotting.plot_svd import plot_rsv_residual
from pyglotaran_extras.plotting.plot_svd import plot_svd
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import add_cycler_if_not_none
from pyglotaran_extras.plotting.utils import extract_irf_location
from pyglotaran_extras.types import Unset

if TYPE_CHECKING:
    from cycler import Cycler
    from glotaran.project.result import Result
    from matplotlib.figure import Figure
    from matplotlib.pyplot import Axes

    from pyglotaran_extras.types import DatasetConvertible
    from pyglotaran_extras.types import UnsetType


@use_plot_config(exclude_from_config=("cycler", "das_cycler", "svd_cycler"))
def plot_overview(
    result: DatasetConvertible | Result,
    center_λ: float | None = None,
    linlog: bool = True,
    linthresh: float = 1,
    linscale: float = 1,
    show_data: bool | None = False,
    main_irf_nr: int = 0,
    figsize: tuple[float, float] = (18, 16),
    cycler: Cycler | None = PlotStyle().cycler,
    figure_only: bool | None = None,
    nr_of_data_svd_vectors: int = 4,
    nr_of_residual_svd_vectors: int = 2,
    show_data_svd_legend: bool = True,
    show_residual_svd_legend: bool = True,
    show_irf_dispersion_center: bool = True,
    show_zero_line: bool = True,
    das_cycler: Cycler | None | UnsetType = Unset,
    svd_cycler: Cycler | None | UnsetType = Unset,
    use_svd_number: bool = False,
) -> tuple[Figure, Axes]:
    """Plot overview of the optimization result.

    Parameters
    ----------
    result : DatasetConvertible | Result
        Result from a pyglotaran optimization as dataset, Path or Result object.
    center_λ : float | None
        Center wavelength (λ in nm)
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    linscale : float
        This allows the linear range (-linthresh to linthresh) to be stretched
        relative to the logarithmic range.
        Its value is the number of decades to use for each half of the linear range.
        For example, when linscale == 1.0 (the default), the space used for the
        positive and negative halves of the linear range will be equal to one
        decade in the logarithmic range. Defaults to 1.
    show_data : bool | None
        Whether to show the input data or residual. If set to ``None`` the plot is skipped
        which improves plotting performance for big datasets. Defaults to False.
    main_irf_nr : int
        Index of the main ``irf`` component when using an ``irf``
        parametrized with multiple peaks. Defaults to 0.
    figsize : tuple[float, float]
        Size of the figure (N, M) in inches. Defaults to (18, 16).
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    figure_only : bool | None
        Deprecated please remove this argument for you function calls. Defaults to None.
    nr_of_data_svd_vectors : int
        Number of data SVD vector to plot. Defaults to 4.
    nr_of_residual_svd_vectors : int
        Number of residual SVD vector to plot. Defaults to 2.
    show_data_svd_legend : bool
        Whether or not to show the data SVD legend. Defaults to True.
    show_residual_svd_legend : bool
        Whether or not to show the residual SVD legend. Defaults to True.
    show_irf_dispersion_center : bool
        Whether to show the the IRF dispersion center as overlay on the residual/data plot.
        Defaults to True.
    show_zero_line : bool
        Whether or not to add a horizontal line at zero to the plots of the spectra.
        Defaults to True.
    das_cycler : Cycler | None | UnsetType
        Plot style cycler to use for DAS plots. Defaults to ``Unset`` which means that the value
        of ``cycler`` is used.
    svd_cycler : Cycler | None | UnsetType
        Plot style cycler to use for SVD plots. Defaults to ``Unset`` which means that the value
        of ``cycler`` is used.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.

    Returns
    -------
    tuple[Figure, Axes]
    """
    res = load_data(result, _stacklevel=3)

    if das_cycler is Unset:
        das_cycler = cycler
    if svd_cycler is Unset:
        svd_cycler = cycler

    if res.coords["time"].to_numpy().size == 1:
        fig, axes = plot_guidance(res)
        if figure_only is not None:
            warn(PyglotaranExtrasApiDeprecationWarning(FIG_ONLY_WARNING), stacklevel=2)
        return fig, axes
    fig, axes = plt.subplots(4, 3, figsize=figsize, constrained_layout=True)

    irf_location = extract_irf_location(res, center_λ, main_irf_nr)

    if center_λ is None:  # center wavelength (λ in nm)
        center_λ = min(res.sizes["spectral"], round(res.sizes["spectral"] / 2))

    # First and second row: concentrations - SAS/EAS - DAS
    plot_concentrations(
        res,
        axes[0, 0],
        center_λ,
        linlog=linlog,
        linthresh=linthresh,
        linscale=linscale,
        main_irf_nr=main_irf_nr,
        cycler=cycler,
    )
    plot_spectra(
        res, axes[0:2, 1:3], cycler=cycler, show_zero_line=show_zero_line, das_cycler=das_cycler
    )
    plot_svd(
        res,
        axes[2:4, 0:3],
        linlog=linlog,
        linthresh=linthresh,
        cycler=svd_cycler,
        nr_of_data_svd_vectors=nr_of_data_svd_vectors,
        nr_of_residual_svd_vectors=nr_of_residual_svd_vectors,
        show_data_svd_legend=show_data_svd_legend,
        show_residual_svd_legend=show_residual_svd_legend,
        irf_location=irf_location,
        use_svd_number=use_svd_number,
    )
    plot_residual(
        res,
        axes[1, 0],
        linlog=linlog,
        linthresh=linthresh,
        show_data=show_data,
        cycler=cycler,
        show_irf_dispersion_center=show_irf_dispersion_center,
        irf_location=irf_location,
    )
    if figure_only is not None:
        warn(PyglotaranExtrasApiDeprecationWarning(FIG_ONLY_WARNING), stacklevel=2)
    return fig, axes


@use_plot_config(exclude_from_config=("cycler", "svd_cycler"))
def plot_simple_overview(
    result: DatasetConvertible | Result,
    title: str | None = None,
    figsize: tuple[float, float] = (12, 6),
    cycler: Cycler | None = PlotStyle().cycler,
    figure_only: bool | None = None,
    show_irf_dispersion_center: bool = True,
    show_data: bool | None = False,
    svd_cycler: Cycler | None | UnsetType = Unset,
    use_svd_number: bool = False,
) -> tuple[Figure, Axes]:
    """Plot simple overview.

    Parameters
    ----------
    result : DatasetConvertible | Result
        Result from a pyglotaran optimization as dataset, Path or Result object.
    title : str | None
        Title of the figure. Defaults to None.
    figsize : tuple[float, float]
        Size of the figure (N, M) in inches. Defaults to (18, 16).
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler.
    figure_only : bool | None
        Deprecated please remove this argument for you function calls. Defaults to None.
    show_irf_dispersion_center : bool
        Whether to show the the IRF dispersion center as overlay on the residual/data plot.
        Defaults to True.
    show_data : bool | None
        Whether to show the input data or residual. If set to ``None`` the plot is skipped
        which improves plotting performance for big datasets. Defaults to False.
    svd_cycler : Cycler | None | UnsetType
        Plot style cycler to use for SVD plots. Defaults to ``Unset`` which means that the value
        of ``cycler`` is used.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.

    Returns
    -------
    tuple[Figure, Axes]
    """
    res = load_data(result, _stacklevel=3)
    if svd_cycler is Unset:
        svd_cycler = cycler

    fig, axes = plt.subplots(2, 3, figsize=figsize, constrained_layout=True)
    for ax in axes.flatten():
        add_cycler_if_not_none(ax, cycler)
    if title:
        fig.suptitle(title, fontsize=16)

    plot_concentrations(res, ax=axes[0, 0], center_λ=res.coords["spectral"].to_numpy()[0])
    plot_sas(res, ax=axes[0, 1])

    irf_location = extract_irf_location(res, center_λ=res.coords["spectral"].to_numpy()[0])

    plot_lsv_residual(
        res,
        ax=axes[1, 0],
        irf_location=irf_location,
        cycler=svd_cycler,
        use_svd_number=use_svd_number,
    )
    plot_rsv_residual(res, ax=axes[1, 1], cycler=svd_cycler, use_svd_number=use_svd_number)

    plot_residual(
        res,
        axes[0, 2],
        show_data=show_data,
        show_irf_dispersion_center=show_irf_dispersion_center,
        irf_location=irf_location,
    )
    plot_residual(
        res,
        axes[1, 2],
        show_data=False,
        show_irf_dispersion_center=show_irf_dispersion_center,
        irf_location=irf_location,
    )

    if figure_only is not None:
        warn(PyglotaranExtrasApiDeprecationWarning(FIG_ONLY_WARNING), stacklevel=2)
    return fig, axes


if __name__ == "__main__":
    import sys

    result_path = Path(sys.argv[1])
    res = load_data(result_path)
    print(res)  # noqa: T201

    fig, axes = plot_overview(res)
    if len(sys.argv) > 2:
        fig.savefig(sys.argv[2], bbox_inches="tight")
        print(f"Saved figure to: {sys.argv[2]}")  # noqa: T201
    else:
        plt.show(block=False)
        input("press <ENTER> to continue")



"""Package containing io convenience functionality."""

from __future__ import annotations

from pyglotaran_extras.io.load_data import load_data
from pyglotaran_extras.io.setup_case_study import setup_case_study

__all__ = ["setup_case_study", "load_data"]



"""Module containing plotting utility functionality."""

from __future__ import annotations

import itertools
from collections.abc import Iterable
from math import ceil
from math import log
from types import MappingProxyType
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
import xarray as xr
from matplotlib.ticker import Locator

from pyglotaran_extras.inspect.utils import pretty_format_numerical_iterable
from pyglotaran_extras.io.utils import result_dataset_mapping

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Hashable
    from collections.abc import Mapping
    from typing import Literal

    from collecations.abs import Sequence
    from cycler import Cycler
    from matplotlib.axis import Axis
    from matplotlib.figure import Figure
    from matplotlib.pyplot import Axes

    from pyglotaran_extras.types import BuiltinSubPlotLabelFormatFunctionKey
    from pyglotaran_extras.types import CyclerColor
    from pyglotaran_extras.types import ResultLike
    from pyglotaran_extras.types import SubPlotLabelCoord


class PlotDuplicationWarning(UserWarning):
    """Warning given when there are more subplots than datapoints."""


def select_irf_dispersion_center_by_index(
    irf_dispersion: xr.DataArray, main_irf_nr: int = 0
) -> xr.DataArray | float:
    """Select a subset of the IRF dispersion data where ``irf_nr==main_irf_nr``.

    Parameters
    ----------
    irf_dispersion : xr.DataArray
        Data Variable from a result dataset which contains the IRF dispersion data.
    main_irf_nr : int
        Index of the main ``irf`` component when using an ``irf``
        parametrized with multiple peaks. Defaults to 0.

    Returns
    -------
    xr.DataArray | float
        DataArray only containing the IRF dispersion data for the main IRF.

    Raises
    ------
    ValueError
        If ``irf_nr`` is not in the coordinates
    """
    if "irf_nr" in irf_dispersion.sizes:
        if main_irf_nr >= irf_dispersion.sizes["irf_nr"]:
            msg = (
                f"The value main_irf_nr={main_irf_nr!r} is not a valid value for irf_nr, needs "
                f"to be smaller than {irf_dispersion.sizes['irf_nr']}."
            )
            raise ValueError(msg)
        irf_dispersion = irf_dispersion.sel(irf_nr=main_irf_nr)
    if irf_dispersion.size == 1:
        irf_dispersion = irf_dispersion.item()
    return irf_dispersion


def extract_irf_dispersion_center(
    res: xr.Dataset, main_irf_nr: int = 0, *, as_dataarray: bool = True
) -> xr.DataArray | float:
    """Extract the IRF dispersion center data from a result dataset where ``irf_nr==irf_nr_index``.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset from a pyglotaran optimization.
    main_irf_nr : int
        Index of the main ``irf`` component when using an ``irf``
        parametrized with multiple peaks. Defaults to 0.
    as_dataarray : bool
        Ensure that the returned data are xr.DataArray instead of a float, even if the
        dispersion is none existent or constant. Defaults to True


    Returns
    -------
    xr.DataArray | float
        IRF dispersion data as float or xr.DataArray
    """
    if "irf_center_location" in res:
        return select_irf_dispersion_center_by_index(
            res.irf_center_location, main_irf_nr=main_irf_nr
        )
    if "center_dispersion_1" in res:
        # legacy compatibility pyglotaran<0.5.0
        return select_irf_dispersion_center_by_index(
            res.center_dispersion_1, main_irf_nr=main_irf_nr
        )

    # No/constant dispersion
    if "irf_center" in res:
        irf_dispersion_center = select_irf_dispersion_center_by_index(
            res.irf_center, main_irf_nr=main_irf_nr
        )
    else:
        irf_dispersion_center = min(res.coords["time"]).item()

    if as_dataarray is True:
        spectral = res.coords["spectral"].to_numpy()
        return xr.DataArray(
            irf_dispersion_center * np.ones(spectral.shape), {"spectral": spectral}
        )
    return irf_dispersion_center


def extract_irf_location(
    res: xr.Dataset, center_λ: float | None = None, main_irf_nr: int | None = 0
) -> float:
    """Determine location of the ``irf``, which can be used to shift plots.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset from a pyglotaran optimization.
    center_λ : float | None
        Center wavelength (λ in nm)
    main_irf_nr : int | None
        Index of the main ``irf`` component when using an ``irf`` parametrized with multiple peaks.
        If it is none ``None`` the location will be 0. Defaults to 0.

    Returns
    -------
    float
        Location of the ``irf``
    """
    if main_irf_nr is None:
        return 0
    irf_dispersion_center = extract_irf_dispersion_center(
        res=res, main_irf_nr=main_irf_nr, as_dataarray=False
    )
    if isinstance(irf_dispersion_center, xr.DataArray):
        if center_λ is None:  # center wavelength (λ in nm)
            center_λ = min(res.sizes["spectral"], round(res.sizes["spectral"] / 2))
        return irf_dispersion_center.sel(spectral=center_λ, method="nearest").item()

    return irf_dispersion_center


def maximum_coordinate_range(
    result: ResultLike, coord_name: str = "spectral"
) -> tuple[float, float]:
    """Calculate the minimal and maximal values of a coordinate across datasets.

    Parameters
    ----------
    result : ResultLike
        Data structure which can be converted to a mapping.
    coord_name : str
        Name of the coordinate to calculate the maximal range for.

    Returns
    -------
    tuple[float, float]
        Minimal and maximal values across datasets

    See Also
    --------
    plot_fit_overview
    """
    result_map = result_dataset_mapping(result)
    minima = []
    maxima = []
    for dataset in result_map.values():
        coord = dataset.coords[coord_name].to_numpy()
        minima.append(coord.min())
        maxima.append(coord.max())
    return min(minima), max(maxima)


def add_unique_figure_legend(fig: Figure, axes: Axes) -> None:
    """Add a legend with unique elements sorted by label to a figure.

    The handles and labels are extracted from the ``axes``

    Parameters
    ----------
    fig : Figure
        Figure to add the legend to.
    axes : Axes
        Axes plotted on the figure.

    See Also
    --------
    plot_fit_overview
    """
    handles = []
    labels = []
    for ax in axes.flatten():
        ax_handles, ax_labels = ax.get_legend_handles_labels()
        handles += ax_handles
        labels += ax_labels
    unique = [
        (handle, label)
        for i, (handle, label) in enumerate(zip(handles, labels, strict=True))
        if label not in labels[:i]
    ]
    unique.sort(key=lambda entry: entry[1])
    fig.legend(*zip(*unique, strict=True))


def get_next_cycler_color(ax: Axes) -> CyclerColor:
    """Get next color from cycler assigned to ``ax``.

    Note
    ----
    This will advance the cycler to the next state.

    Parameters
    ----------
    ax : Axes
        Axes to get the color from.

    Returns
    -------
    CyclerColor
    """
    # Matplotlib<3.8 compat
    if hasattr(ax._get_lines, "prop_cycler"):
        return next(ax._get_lines.prop_cycler)
    return {"color": ax._get_lines.get_next_color()}


def select_plot_wavelengths(
    result: ResultLike,
    axes_shape: tuple[int, int] = (4, 4),
    wavelength_range: tuple[float, float] | None = None,
    equidistant_wavelengths: bool = True,
) -> Iterable[float]:
    """Select wavelengths to be used in ``plot_fit_overview`` from a result.

    Parameters
    ----------
    result : ResultLike
        Data structure which can be converted to a mapping of datasets.
    axes_shape : tuple[int, int]
        Shape of the plot grid (N, M). Defaults to (4, 4).
    wavelength_range : tuple[float, float] | None
        Tuple of minimum and maximum values to calculate the the wavelengths
        used for plotting. If not provided the values will be tetermined over all datasets.
        Defaults to None.
    equidistant_wavelengths : bool
        Whether or not wavelengths should be selected based on equidistant values
        or equidistant indices (only supported for a single dataset).
        Since in general multiple datasets will have. Defaults to True.

    Returns
    -------
    Iterable[float]
        Wavelength which should be used for each subplot by ``plot_fit_overview``.

    See Also
    --------
    maximum_coordinate_range
    """
    result_map = result_dataset_mapping(result)
    nr_of_plots = np.prod(axes_shape)

    if wavelength_range is None:
        wavelength_range = maximum_coordinate_range(result_map)

    if equidistant_wavelengths:
        return np.linspace(*wavelength_range, num=nr_of_plots)

    first_dataset = next(iter(result_map.keys()))
    if len(result_map) > 1:
        warn(
            UserWarning(
                "Calculating plot wavelengths is only supported, for a single dataset."
                f"The dataset {first_dataset!r}, will be used to calculate the selected "
                "wavelengths.To mute this warning call "
                f"'{select_plot_wavelengths.__name__}' with only one dataset."
            ),
            stacklevel=2,
        )
    spectral_coords = result_map[first_dataset].coords["spectral"]
    spectral_coords = spectral_coords[
        (wavelength_range[0] <= spectral_coords) & (spectral_coords <= wavelength_range[1])
    ]
    spectral_indices = np.linspace(0, len(spectral_coords) - 1, num=nr_of_plots, dtype=np.int64)
    return spectral_coords[spectral_indices].to_numpy()


def extract_dataset_scale(res: xr.Dataset, divide_by_scale: bool = True) -> float:
    """Extract 'dataset_scale' attribute from optimization result dataset.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset from a pyglotaran optimization.
    divide_by_scale : bool
        Whether or not to divide the data by the dataset scale used for optimization.
        Defaults to True.

    Returns
    -------
    float
        Dataset scale extracted from ``res`` falls back to 1 if not present.
    """
    scale = 1
    if divide_by_scale is True:
        if "dataset_scale" in res.attrs:
            scale = res.attrs["dataset_scale"]
        else:
            warn(
                UserWarning(
                    "Diving data by dataset scales is only supported by results from "
                    "'pyglotaran>=0.5.0'. Please upgrade pyglotaran and recreate the "
                    "result to plot."
                ),
                stacklevel=2,
            )
    return scale


def shift_time_axis_by_irf_location(
    plot_data: xr.DataArray, irf_location: float | None, *, _internal_call: bool = False
) -> xr.DataArray:
    """Shift ``plot_data`` 'time' axis  by the position of the main ``irf``.

    Parameters
    ----------
    plot_data : xr.DataArray
        Data to plot.
    irf_location : float | None
        Location of the ``irf``, if the value is None the original ``plot_data`` will be returned.
    _internal_call : bool
        This indicates internal use stripping away user help and silently skipping execution.
        Defaults to False.

    Returns
    -------
    xr.DataArray
        ``plot_data`` with the time axis shifted by the position of the main ``irf``.

    Raises
    ------
    ValueError
        If ``plot_data`` does not have a time axis.

    See Also
    --------
    extract_irf_location
    """
    if irf_location is None or ("time" not in plot_data.coords and _internal_call is True):
        return plot_data

    if "time" not in plot_data.coords:
        msg = "plot_data need to have a 'time' axis."
        raise ValueError(msg)

    times_shifted = plot_data.coords["time"] - irf_location
    return plot_data.assign_coords(time=times_shifted)


def get_shifted_traces(
    res: xr.Dataset, center_λ: float | None = None, main_irf_nr: int = 0
) -> xr.DataArray:
    """Shift traces by the position of the main ``irf``.

    Parameters
    ----------
    res : xr.Dataset
        Result dataset from a pyglotaran optimization.
    center_λ : float | None
        Center wavelength (λ in nm). Defaults to None.
    main_irf_nr : int
        Index of the main ``irf`` component when using an ``irf``
        parametrized with multiple peaks. Defaults to 0.

    Returns
    -------
    xr.DataArray
        Traces shifted by the ``irf``s location, to align the at 0.

    Raises
    ------
    ValueError
        If no known concentration was found in the result.
    """
    if "species_concentration" in res:
        traces = res.species_concentration
    elif "species_associated_concentrations" in res:
        traces = res.species_associated_concentrations
    else:
        msg = f"No concentrations in result:\n{res}"
        raise ValueError(msg)

    irf_location = extract_irf_location(res, center_λ, main_irf_nr)
    return shift_time_axis_by_irf_location(traces, irf_location)


def ensure_axes_array(axes: Axis | Axes) -> Axes:
    """Ensure that axes have flatten method even if it is a single axis.

    Parameters
    ----------
    axes : Axis | Axes
        Axis or Axes to convert for API consistency.

    Returns
    -------
    Axes
        Numpy ndarray of axes.
    """
    # We can't use `Axis` in isinstance so we check for the np.ndarray attribute of `Axes`
    if hasattr(axes, "flatten") is False:
        axes = np.array([axes])
    return axes


def add_cycler_if_not_none(axis: Axis | Axes, cycler: Cycler | None) -> None:
    """Add cycler to and axis if it is not None.

    This is a convenience function that allow to opt out of using
    a cycler, which is needed to run a plotting function in a loop
    where the cycler is controlled from the outside.


    Parameters
    ----------
    axis : Axis | Axes
        Axis to plot on.
    cycler : Cycler | None
        Plot style cycler to use.
    """
    if cycler is not None:
        axis = ensure_axes_array(axis)
        for ax in axis.flatten():
            ax.set_prop_cycle(cycler)


def abs_max(
    data: xr.DataArray, *, result_dims: Hashable | Iterable[Hashable] = ()
) -> xr.DataArray:
    """Calculate the absolute maximum values of ``data`` along all dims except ``result_dims``.

    Parameters
    ----------
    data : xr.DataArray
        Data for which the absolute maximum should be calculated.
    result_dims : Hashable | Iterable[Hashable]
        Dimensions of ``data`` which should be preserved and part of the resulting DataArray.
        Defaults to () which results in using the absolute maximum of all values.

    Returns
    -------
    xr.DataArray
        Absolute maximum values of ``data`` with dimensions ``result_dims``.
    """
    if not isinstance(result_dims, Iterable):
        result_dims = (result_dims,)
    reduce_dims = (dim for dim in data.dims if dim not in result_dims)
    return np.abs(data).max(dim=reduce_dims)


def calculate_ticks_in_units_of_pi(
    values: np.ndarray | xr.DataArray, *, step_size: float = 0.5
) -> tuple[Iterable[float], Iterable[str]]:
    r"""Calculate tick values and labels in units of Pi.

    Parameters
    ----------
    values : np.ndarray | xr.DataArray
        Values which the ticks should be calculated for.
    step_size : float
        Step size of the ticks in units of pi. Defaults to 0.5

    Returns
    -------
    tuple[Iterable[float], Iterable[str]]
        Tick values and tick labels

    See Also
    --------
    pyglotaran_extras.plotting.plot_doas.plot_doas

    Examples
    --------
    If you have a case study that uses a ``damped-oscillation`` megacomplex you can plot the
    ``damped_oscillation_phase`` with y-tick in units of Pi by the following code given that the
    dataset is saved under ``dataset.nc``.

    .. code-block:: python

        import matplotlib.pyplot as plt

        from glotaran.io import load_dataset
        from pyglotaran_extras.plotting.utils import calculate_ticks_in_units_of_pi

        dataset = load_dataset("dataset.nc")

        fig, ax = plt.subplots(1, 1)

        damped_oscillation_phase = dataset["damped_oscillation_phase"].sel(
            damped_oscillation=["osc1"]
        )
        damped_oscillation_phase.plot.line(x="spectral", ax=ax)

        ax.set_yticks(
            *calculate_ticks_in_units_of_pi(damped_oscillation_phase), rotation="horizontal"
        )
    """
    values = np.array(values)
    int_values_over_pi = np.round(values / np.pi / step_size)
    tick_labels = np.arange(int_values_over_pi.min(), int_values_over_pi.max() + 1) * step_size
    return tick_labels * np.pi, (
        str(val) for val in pretty_format_numerical_iterable(tick_labels, decimal_places=1)
    )


def not_single_element_dims(data_array: xr.DataArray) -> list[Hashable]:
    """Names of dimensions in ``data`` which don't have a size equal to one.

    This helper function is for example used to determine if a data only have a single trace,
    since this requires different plotting code (e.g. ``data_array.plot.line(x="time")``).

    Parameters
    ----------
    data_array : xr.DataArray
        DataArray to check if it has only a single dimension.

    Returns
    -------
    list[Hashable]
        Names of dimensions in ``data`` which don't have a size equal to one.
    """
    return [dim for dim, values in data_array.coords.items() if values.size != 1]


class MinorSymLogLocator(Locator):
    """Dynamically find minor tick positions based on major ticks for a symlog scaling.

    Ref.: https://stackoverflow.com/a/45696768
    """

    def __init__(self, linthresh: float, nints: int = 10) -> None:
        """Ticks will be placed between the major ticks.

        The placement is linear for x between -linthresh and linthresh,
        otherwise its logarithmically. nints gives the number of
        intervals that will be bounded by the minor ticks.

        Parameters
        ----------
        linthresh : float
            A single float which defines the range (-x, x), within which the plot is linear.
        nints : int
            Number of minor tick between major ticks. Defaults to 10
        """
        self.linthresh = linthresh
        self.nintervals = nints

    def __call__(self) -> list[float]:
        """Return the locations of the ticks.

        Returns
        -------
        list[float]
            Minor ticks position.
        """
        # Return the locations of the ticks
        majorlocs = self.axis.get_majorticklocs()

        if len(majorlocs) == 1:
            return self.raise_if_exceeds(np.array([]))

        # add temporary major tick locs at either end of the current range
        # to fill in minor tick gaps
        dmlower = majorlocs[1] - majorlocs[0]  # major tick difference at lower end
        dmupper = majorlocs[-1] - majorlocs[-2]  # major tick difference at upper end

        # add temporary major tick location at the lower end
        if majorlocs[0] != 0.0 and (
            (majorlocs[0] != self.linthresh and dmlower > self.linthresh)
            or (dmlower == self.linthresh and majorlocs[0] < 0)
        ):
            majorlocs = np.insert(majorlocs, 0, majorlocs[0] * 10.0)
        else:
            majorlocs = np.insert(majorlocs, 0, majorlocs[0] - self.linthresh)

        # add temporary major tick location at the upper end
        if majorlocs[-1] != 0.0 and (
            (np.abs(majorlocs[-1]) != self.linthresh and dmupper > self.linthresh)
            or (dmupper == self.linthresh and majorlocs[-1] > 0)
        ):
            majorlocs = np.append(majorlocs, majorlocs[-1] * 10.0)
        else:
            majorlocs = np.append(majorlocs, majorlocs[-1] + self.linthresh)

        # iterate through minor locs
        minorlocs: list[float] = []

        # handle the lowest part
        for i in range(1, len(majorlocs)):
            majorstep = majorlocs[i] - majorlocs[i - 1]
            if abs(majorlocs[i - 1] + majorstep / 2) < self.linthresh:
                ndivs = self.nintervals
            else:
                ndivs = self.nintervals - 1

            minorstep = majorstep / ndivs
            locs = np.arange(majorlocs[i - 1], majorlocs[i], minorstep)[1:]
            minorlocs.extend(locs)

        return self.raise_if_exceeds(np.array(minorlocs))

    def tick_values(self, _vmin: float, _vmax: float) -> None:
        """Return the values of the located ticks given **_vmin** and **_vmax** (not implemented).

        Parameters
        ----------
        _vmin : float
            Minimum value.
        _vmax : float
            Maximum value.

        Raises
        ------
        NotImplementedError
            Not used
        """
        msg = f"Cannot get tick locations for a {type(self)} type."
        raise NotImplementedError(msg)


def format_sub_plot_number_upper_case_letter(sub_plot_number: int, size: None | int = None) -> str:
    """Format ``sub_plot_number`` into an upper case letter, that can be used as label.

    Parameters
    ----------
    sub_plot_number : int
        Number of the subplot starting at One.
    size : None | int
        Size of the axes array (number of plots). Defaults to None

    Returns
    -------
    str
        Upper case label for a sub plot.

    Examples
    --------
    >>> print(format_sub_plot_number_upper_case_letter(1))
    A

    >>> print(format_sub_plot_number_upper_case_letter(26))
    Z

    >>> print(format_sub_plot_number_upper_case_letter(27))
    AA

    >>> print(format_sub_plot_number_upper_case_letter(1, 26))
    AA

    >>> print(format_sub_plot_number_upper_case_letter(2, 26))
    AB

    >>> print(format_sub_plot_number_upper_case_letter(26, 26))
    AZ

    >>> print(format_sub_plot_number_upper_case_letter(27, 50))
    BA

    See Also
    --------
    BuiltinLabelFormatFunctions
    add_subplot_labels
    """
    sub_plot_number -= 1
    if size is not None and size > 26:
        return "".join(
            format_sub_plot_number_upper_case_letter(((sub_plot_number // (26**i)) % 26) + 1)
            for i in reversed(range(1, ceil(log(size, 26))))
        ) + format_sub_plot_number_upper_case_letter((sub_plot_number % 26) + 1)
    if sub_plot_number < 26:
        return chr(ord("A") + sub_plot_number)
    return format_sub_plot_number_upper_case_letter(
        sub_plot_number // 26
    ) + format_sub_plot_number_upper_case_letter((sub_plot_number % 26) + 1)


BuiltinSubPlotLabelFormatFunctions: Mapping[str, Callable[[int, int | None], str]] = (
    MappingProxyType(
        {
            "number": lambda x, _: f"{x}",
            "upper_case_letter": format_sub_plot_number_upper_case_letter,
            "lower_case_letter": lambda x, y: format_sub_plot_number_upper_case_letter(
                x, y
            ).lower(),
        }
    )
)


def get_subplot_label_format_function(
    format_function: BuiltinSubPlotLabelFormatFunctionKey | Callable[[int, int | None], str],
) -> Callable[[int, int | None], str]:
    """Get subplot label function from ``BuiltinSubPlotLabelFormatFunctions`` if it is a key.

    This function is mainly needed for typing reasons.

    Parameters
    ----------
    format_function : BuiltinSubPlotLabelFormatFunctionKey | Callable[[int, int | None], str]
        Key ``BuiltinSubPlotLabelFormatFunctions`` to retrieve builtin function or user defined
        format function.

    Returns
    -------
    Callable[[int, int | None], str]
        Function to format subplot label.
    """
    if isinstance(format_function, str) and format_function in BuiltinSubPlotLabelFormatFunctions:
        return BuiltinSubPlotLabelFormatFunctions[format_function]
    return format_function  # type:ignore[return-value]


def add_subplot_labels(
    axes: Axis | Axes,
    *,
    label_position: tuple[float, float] = (-0.05, 1.05),
    label_coords: SubPlotLabelCoord = "axes fraction",
    direction: Literal["row", "column"] = "row",
    label_format_template: str = "{}",
    label_format_function: BuiltinSubPlotLabelFormatFunctionKey
    | Callable[[int, int | None], str] = "number",
    fontsize: int = 16,
) -> None:
    """Add labels to all subplots in ``axes`` in a consistent manner.

    Parameters
    ----------
    axes : Axis | Axes
        Axes (subplots) on which the labels should be added.
    label_position : tuple[float, float]
        Position of the label in ``label_coords`` coordinates.
    label_coords : SubPlotLabelCoord
        Coordinate system used for ``label_position``. Defaults to "axes fraction"
    direction : Literal["row", "column"]
        Direct in which the axes should be iterated in. Defaults to "row"
    label_format_template : str
        Template string to inject the return value of ``label_format_function`` into.
        Defaults to "{}"
    label_format_function : BuiltinSubPlotLabelFormatFunctionKey | Callable[[int, int | None], str]
        Function to calculate the label for the axis index and ``axes`` size. Defaults to "number"
    fontsize : int
        Font size used for the label. Defaults to 16
    """
    axes = ensure_axes_array(axes)
    format_function = get_subplot_label_format_function(label_format_function)
    if direction == "column":
        axes = axes.T
    for i, ax in enumerate(axes.flatten(), start=1):
        ax.annotate(
            label_format_template.format(format_function(i, axes.size)),
            xy=label_position,
            xycoords=label_coords,
            fontsize=fontsize,
        )


def condense_numbered_strings(inputs: Sequence[str] | None) -> str:
    """Condense a list of numbered strings into a compact string representation.

    This function takes a sequence of strings with a common prefix followed by numbers,
    and condenses consecutive numbers into ranges. For example, the input
    ["osc1", "osc2", "osc3", "osc5"] will be condensed to "osc1-3,5".

    Parameters
    ----------
    inputs : Sequence[str] | None
        A sequence of strings to condense. If None or empty, an empty string is returned.

    Returns
    -------
    str
        A condensed string representation of the input sequence.
    """
    if inputs is None:
        return ""

    # Extract prefix and numbers
    prefix = "".join(c for c in inputs[0] if not c.isdigit())
    nums = sorted(int("".join(c for c in s if c.isdigit())) for s in inputs)

    # Find and format ranges
    ranges = []
    for _, group in itertools.groupby(enumerate(nums), key=lambda x: x[0] - x[1]):
        group_list = list(group)
        start = group_list[0][1]
        end = group_list[-1][1]
        if start == end:
            ranges.append(str(start))
        else:
            ranges.append(f"{start}-{end}")

    return f"{prefix}{','.join(ranges)}"



"""Package containing plotting functionality."""

from __future__ import annotations

from pyglotaran_extras.plotting.plot_data import plot_data_overview
from pyglotaran_extras.plotting.plot_overview import plot_overview
from pyglotaran_extras.plotting.plot_overview import plot_simple_overview

__all__ = ["plot_data_overview", "plot_overview", "plot_simple_overview"]



"""Module containing PFID (Perturbed Free Induction Decay) plotting functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np

from pyglotaran_extras.io.load_data import load_data
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import abs_max
from pyglotaran_extras.plotting.utils import add_cycler_if_not_none
from pyglotaran_extras.plotting.utils import calculate_ticks_in_units_of_pi
from pyglotaran_extras.plotting.utils import condense_numbered_strings
from pyglotaran_extras.plotting.utils import extract_irf_location
from pyglotaran_extras.plotting.utils import shift_time_axis_by_irf_location

if TYPE_CHECKING:
    from cycler import Cycler
    from glotaran.project.result import Result
    from matplotlib.figure import Figure
    from matplotlib.pyplot import Axes

    from pyglotaran_extras.types import DatasetConvertible


def plot_pfid(  # noqa: C901
    dataset: DatasetConvertible | Result,
    *,
    names: list[str] | None = None,
    oscillation_type: Literal["cos", "sin"] = "cos",
    show_clps: bool = False,
    time_range: tuple[float, float] | None = None,
    spectral: float | None = None,
    main_irf_nr: int | None = 0,
    normalize: bool = False,
    figsize: tuple[float, float] | None = None,
    show_zero_line: bool = True,
    cycler: Cycler | None = PlotStyle().cycler,
    title: str | None = "Perturbed Free Induction Decays",
    legend_format_string: str = r"{label}: $\nu$={frequency:.0f}, $\gamma$={rate:.1f}",
) -> tuple[Figure, Axes]:
    r"""Plot PFID (Perturbed Free Induction Decay) related data of the optimization result.

    Parameters
    ----------
    dataset : DatasetConvertible | Result
        Result dataset from a pyglotaran optimization.
    names : list[str] | None
        List of oscillation names which should be plotted.
        Defaults to None which means that all oscillations will be plotted.
    oscillation_type : Literal["cos", "sin"]
        Type of the oscillation to show in the oscillation plot. Defaults to "cos"
    show_clps : bool, optional
        If True, plots additional conditional linear parameters (CLPs), in which case the
        oscillation_type argument is ignored (both cos and sin are plotted). Defaults to False.
    time_range : tuple[float, float] | None
        Start and end time for the Oscillation plot, if ``main_irf_nr`` is not None the value are
        relative to the IRF location. Defaults to None which means that the full time range is
        used.
    spectral : float | None
        Value of the spectral axis that should be used to select the data for the Oscillation
        plot this value does not need to be an exact existing value and only has effect if the
        IRF has dispersion. If None the Oscillation plot at lowest spectral value will be shown.
    main_irf_nr : int | None
        Index of the main ``irf`` component when using an ``irf`` parametrized with multiple peaks
        and is used to shift the time axis. If it is none ``None`` the shifting will be
        deactivated. Defaults to 0.
    normalize : bool
        Whether or not to normalize the PFID spectra plot. If the PFID spectra is normalized,
        the Oscillation is scaled with the reciprocal of the normalization to compensate for this.
        Defaults to False.
    figsize : tuple[float, float] | None
        Size of the figure (N, M) in inches. Defaults to None which then uses
        (20,5) if show_clps=False, (20, 10) if show_clps=True.
    show_zero_line : bool
        Whether or not to add a horizontal line at zero. Defaults to True
    cycler : Cycler | None
        Plot style cycler to use. Defaults to PlotStyle().cycler
    title : str | None
        Title of the figure. Defaults to "Perturbed Free Induction Decays"
    legend_format_string : str
        Format string for each entry in the legend of the oscillation plot. Possible values which
        can be replaced are ``label`` (label of the oscillation in the model definition),
        ``frequency`` (ν) and ``rate`` (γ). Use ``""`` to remove the legend. Defaults to
        ``r"{label}: $\nu$={frequency:.0f}, $\gamma$={rate:.1f}"``

    Returns
    -------
    tuple[Figure, Axes]
        Figure object which contains the plots and the Axes.

    See Also
    --------
    calculate_ticks_in_units_of_pi
    """
    dataset = load_data(dataset, _stacklevel=3)
    if figsize is None:
        figsize = (20, 10) if show_clps else (20, 5)
    fig, axes = plt.subplots(2 if show_clps else 1, 3, figsize=figsize)

    add_cycler_if_not_none(axes, cycler)

    time_sel_kwargs = {"time": slice(time_range[0], time_range[1])} if time_range else {}
    names = dataset.pfid.to_numpy() if names is None else names
    osc_sel_kwargs = {"pfid": names}
    if spectral is None and "spectral" in dataset.coords:
        spectral = dataset.coords["spectral"].min().item()
    irf_location = extract_irf_location(dataset, spectral, main_irf_nr)

    pfid_phase = dataset["pfid_phase"].sel(**osc_sel_kwargs)
    oscillations_spectra = dataset["pfid_associated_spectra"].sel(**osc_sel_kwargs)
    oscillation_types = ["cos", "sin"] if show_clps else [oscillation_type]

    for axes_index, osc_type in enumerate(oscillation_types):
        oscillations = dataset[f"pfid_{osc_type}"]

        if "spectral" in oscillations.coords:
            oscillations = oscillations.sel(spectral=spectral, method="nearest")

        oscillations = shift_time_axis_by_irf_location(
            oscillations.sel(**osc_sel_kwargs), irf_location, _internal_call=True
        )
        osc_max = abs_max(oscillations, result_dims="pfid")
        spectra_max = abs_max(oscillations_spectra, result_dims="pfid")
        scales = np.sqrt(osc_max * spectra_max)
        norm_factor = scales.max() if normalize else 1

        oscillations_to_plot = (oscillations / osc_max * scales * norm_factor).sel(
            **time_sel_kwargs
        )

        for oscillation_label in oscillations_to_plot.pfid.to_numpy():
            oscillation = oscillations_to_plot.sel(pfid=[oscillation_label])
            frequency = oscillation.pfid_frequency.item()
            rate = oscillation.pfid_rate.item()
            oscillation.plot.line(
                x="time",
                ax=axes[axes_index, 0] if show_clps else axes[0],
                label=legend_format_string.format(
                    label=oscillation_label, frequency=frequency, rate=rate
                ),
            )

    spectra_ax = axes[0, 2] if show_clps else axes[1]
    phases_ax = axes[1, 2] if show_clps else axes[2]

    (oscillations_spectra / spectra_max * scales / norm_factor).plot.line(
        x="spectral", ax=spectra_ax
    )
    pfid_phase.plot.line(x="spectral", ax=phases_ax)

    spectra_ax.set_title("Spectra")
    spectra_ax.set_ylabel("Normalized PFID" if normalize else "PFID")

    phases_ax.set_title("Phases")
    phases_ax.set_yticks(*calculate_ticks_in_units_of_pi(pfid_phase), rotation="horizontal")
    phases_ax.set_ylabel("Phase (π)")

    if show_clps:
        for axes_index, osc_type in enumerate(oscillation_types):
            for pfidname in names:
                clp_label = f"{pfidname}_{osc_type}"
                dataset["clp"].sel(clp_label=clp_label).plot(ax=axes[axes_index, 1])
            axes[axes_index, 1].set_title(f"clps={condense_numbered_strings(names)}")

        axes[0, 0].set_title(f"Cos Oscillations {spectral}")
        axes[1, 0].set_title(f"Sin Oscillations {spectral}")
    else:
        axes[0].set_title(f"{oscillation_type.capitalize()} Oscillations {spectral}")

    for ax in axes.flatten():
        legend = ax.get_legend()
        if legend is not None:
            legend.remove()

    if legend_format_string:
        axes[0].legend() if not show_clps else axes[0, 0].legend()

    if show_zero_line:
        [ax.axhline(0, color="k", linewidth=1) for ax in axes.flatten()]

    if title:
        fig.suptitle(title, fontsize=16)

    fig.tight_layout()
    return fig, axes



"""Module containing plot configuration."""

from __future__ import annotations

from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import MutableMapping
from contextlib import contextmanager
from functools import wraps
from inspect import Parameter
from inspect import getcallargs
from inspect import signature
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import TypeAlias
from typing import TypedDict
from typing import cast

import numpy as np
from docstring_parser import parse as parse_docstring
from matplotlib.axes import Axes
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import RootModel
from pydantic import ValidationError
from pydantic import field_validator
from pydantic import model_serializer
from pydantic import model_validator
from pydantic_core import ErrorDetails
from pydantic_core import PydanticUndefined

from pyglotaran_extras.config.utils import add_yaml_repr

if TYPE_CHECKING:
    from collections.abc import Callable

    from pyglotaran_extras.config.config import Config
    from pyglotaran_extras.types import Param
    from pyglotaran_extras.types import RetType


class DefaultKwarg(TypedDict):
    """Default value and type annotation of a kwarg extracted from the function signature."""

    default: Any
    annotation: str
    docstring: str | None


DefaultKwargs: TypeAlias = Mapping[str, DefaultKwarg]
__PlotFunctionRegistry: MutableMapping[str, DefaultKwargs] = {}


@add_yaml_repr
class PlotLabelOverrideValue(BaseModel):
    """Value of ``PlotLabelOverrideMap``."""

    model_config = ConfigDict(extra="forbid")

    target_name: str
    axis: Literal["x", "y", "both"] = "both"

    @model_serializer
    def serialize(self) -> dict[str, Any] | str:
        """Serialize supporting short notation.

        Returns
        -------
        dict[str, Any] | str
        """
        if self.axis == "both":
            return self.target_name
        return {"target_name": self.target_name, "axis": self.axis}


def _add_short_notation_to_schema(json_schema: dict[str, Any]) -> None:  # noqa: DOC
    """Update json schema to support short notation for ``PlotLabelOverrideValue``."""
    orig_additional_properties = json_schema["additionalProperties"]
    json_schema["additionalProperties"] = {
        "anyOf": [orig_additional_properties, {"type": "string"}]
    }


@add_yaml_repr
class PlotLabelOverrideMap(RootModel, Mapping):
    """Mapping to override axis labels."""

    model_config = ConfigDict(json_schema_extra=_add_short_notation_to_schema)

    root: dict[str, PlotLabelOverrideValue] = Field(default_factory=dict)

    @model_validator(mode="before")
    @classmethod
    def parse(cls, values: dict[str, Any]) -> dict[str, PlotLabelOverrideValue]:  # noqa: DOC
        """Parse ``axis_label_override`` dictionary supporting verbose and short notation.

        Parameters
        ----------
        values : dict[str, Any]
            Dict that initializes the class.

        Returns
        -------
        dict[str, PlotLabelOverrideValue]
        """
        if values is PydanticUndefined or values is None:
            return {}
        errors: dict[str, ErrorDetails] = {}
        parsed_values: dict[str, PlotLabelOverrideValue] = {}
        for key, value in values.items():
            try:
                if isinstance(value, str):
                    parsed_values[key] = PlotLabelOverrideValue(target_name=value)
                else:
                    parsed_values[key] = PlotLabelOverrideValue.model_validate(value)
            except ValidationError as error:
                errors |= {str(e): e for e in error.errors()}
        if len(errors) > 0:
            raise ValidationError.from_exception_data(cls.__name__, line_errors=[*errors.values()])  # type:ignore[list-item]
        return parsed_values

    def __iter__(self) -> Iterator[str]:  # type:ignore[override] # noqa: DOC
        """Iterate over items."""
        return iter(self.root)

    def __len__(self) -> int:  # noqa: DOC
        """Get number of items."""
        return len(self.root)

    def __getitem__(self, item_label: str) -> PlotLabelOverrideValue:  # noqa: DOC
        """Access items."""
        return self.root[item_label]

    def __contains__(self, item_label: object) -> bool:  # noqa: DOC
        """Check if item is ``in`` the object."""
        return item_label in self.root

    def find_axis_label(self, matplotlib_label: str, axis_name: Literal["x", "y"]) -> str | None:
        """Find axis label even if ``matplotlib`` or the user added a newline in it.

        Parameters
        ----------
        matplotlib_label : str
            Label extracted from the ``matplotlib`` ``Axes`` with ``ax.get_xlabel()`` or
            ``ax.get_xlabel()``.
        axis_name : Literal["x", "y"]
            Name of the axis to find the label for.

        Returns
        -------
        str | None
            Mapped label value if found and None otherwise.
        """
        if matplotlib_label in self and self[matplotlib_label].axis in (axis_name, "both"):
            return self[matplotlib_label].target_name

        # If a label is too long to fit matplotlib inserts a newline which means we can not look it
        # up with string equality
        for key, value in self.root.items():
            if matplotlib_label.replace("\n", "") == key.replace("\n", "") and value.axis in (
                axis_name,
                "both",
            ):
                return value.target_name
        return None


@add_yaml_repr
class PerFunctionPlotConfig(BaseModel):
    """Per function plot configuration."""

    model_config = ConfigDict(extra="forbid")

    default_args_override: dict[str, Any] = Field(
        default_factory=dict,
        description="Default arguments to use if not specified in function call.",
    )
    axis_label_override: PlotLabelOverrideMap | dict[str, str] = Field(
        default_factory=PlotLabelOverrideMap
    )

    @field_validator("axis_label_override", mode="before")
    @classmethod
    def validate_axis_label_override(  # noqa: DOC
        cls, value: PlotLabelOverrideMap | dict[str, str]
    ) -> PlotLabelOverrideMap:
        """Ensure that ``axis_label_override`` gets converted into ``PlotLabelOverrideMap``."""
        return PlotLabelOverrideMap.model_validate(value)

    @model_serializer
    def serialize(self) -> dict[str, Any]:
        """Serialize in a sparse manner leaving out empty values.

        Returns
        -------
        dict[str, Any]
        """
        serialized = {}
        if len(self.default_args_override) > 0:
            serialized["default_args_override"] = self.default_args_override
        if len(self.axis_label_override) > 0:
            serialized["axis_label_override"] = cast(
                PlotLabelOverrideMap, self.axis_label_override
            ).model_dump()
        return serialized

    def merge(self, other: PerFunctionPlotConfig) -> PerFunctionPlotConfig:
        """Merge two ``PerFunctionPlotConfig``'s where ``other`` overrides values.

        Parameters
        ----------
        other : PerFunctionPlotConfig
            Other ``PerFunctionPlotConfig`` to merge in.

        Returns
        -------
        PerFunctionPlotConfig
        """
        self_dict = self.model_dump()
        other_dict = other.model_dump()
        return PerFunctionPlotConfig.model_validate(
            {
                "default_args_override": (
                    self_dict.pop("default_args_override", {})
                    | other_dict.pop("default_args_override", {})
                ),
                "axis_label_override": (
                    self_dict.pop("axis_label_override", {})
                    | other_dict.pop("axis_label_override", {})
                ),
            }
        )

    def find_override_kwargs(self, not_user_provided_kwargs: set[str]) -> dict[str, Any]:
        """Config key word arguments that were not provided by the user and are safe to override.

        Parameters
        ----------
        not_user_provided_kwargs : set[str]
            Set of keyword arguments that were provided by the user and thus should not be
            overridden.

        Returns
        -------
        dict[str, Any]
        """
        return {
            k: self.default_args_override[k]
            for k in self.default_args_override
            if k in not_user_provided_kwargs
        }

    def update_axes_labels(self, axes: Axes | Iterable[Axes]) -> None:
        """Apply label overrides to ``axes``.

        Parameters
        ----------
        axes : Axes | Iterable[Axes]
            Axes to apply the override to.
        """
        if isinstance(axes, Axes):
            self.update_axes_labels((axes,))
            return
        for ax in axes:
            if isinstance(ax, Axes):
                orig_x_label = ax.get_xlabel()
                orig_y_label = ax.get_ylabel()
                axis_label_override = cast(PlotLabelOverrideMap, self.axis_label_override)

                if (
                    override_label := axis_label_override.find_axis_label(orig_x_label, "x")
                ) is not None:
                    ax.set_xlabel(override_label)

                if (
                    override_label := axis_label_override.find_axis_label(orig_y_label, "y")
                ) is not None:
                    ax.set_ylabel(override_label)

            elif isinstance(ax, np.ndarray):
                self.update_axes_labels(ax.flatten())
            else:
                self.update_axes_labels(ax)


@add_yaml_repr
class PlotConfig(BaseModel):
    """Config for plot functions including default args and label overrides."""

    model_config = ConfigDict(extra="allow")

    general: PerFunctionPlotConfig = Field(
        default_factory=PerFunctionPlotConfig,
        description="Config that gets applied to all functions if not specified otherwise.",
    )

    @model_validator(mode="before")
    @classmethod
    def parse(cls, values: dict[str, Any]) -> dict[str, PerFunctionPlotConfig]:
        """Ensure the extra values are converted to ``PerFunctionPlotConfig``.

        Parameters
        ----------
        values : dict[str, Any]
            Dict that initializes the class.

        Returns
        -------
        dict[str, PerFunctionPlotConfig]

        Raises
        ------
        ValidationError
        """
        parsed_values = {}
        errors: dict[str, ErrorDetails] = {}
        for key, value in values.items():
            try:
                parsed_values[key] = PerFunctionPlotConfig.model_validate(value)
            except ValidationError as error:
                errors |= {str(e): {**e, "loc": (key, *e["loc"])} for e in error.errors()}
        if len(errors) > 0:
            raise ValidationError.from_exception_data(cls.__name__, line_errors=[*errors.values()])  # type:ignore[list-item]
        return parsed_values

    def get_function_config(self, function_name: str) -> PerFunctionPlotConfig:
        """Get config for a specific function.

        Parameters
        ----------
        function_name : str
            Name of the function to get the config for.

        Returns
        -------
        PerFunctionPlotConfig
        """
        function_config = self.general
        if self.model_extra is not None and function_name in self.model_extra:
            function_config = function_config.merge(self.model_extra[function_name])
        if hasattr(self, "__context_config"):
            function_config = function_config.merge(getattr(self, "__context_config"))
        return function_config

    def merge(self, other: PlotConfig) -> PlotConfig:  # noqa: C901
        """Merge two ``PlotConfig``'s where ``other`` overrides values.

        Parameters
        ----------
        other : PlotConfig
            Other ``PlotConfig`` to merge in.

        Returns
        -------
        PlotConfig
        """
        updated: dict[str, PerFunctionPlotConfig] = {}
        # Update general field
        for key in self.model_fields_set:
            updated[key] = cast(PerFunctionPlotConfig, getattr(self, key))
            if key in other.model_fields_set:
                updated[key] = updated[key].merge(cast(PerFunctionPlotConfig, getattr(other, key)))
        for key in other.model_fields_set:
            if key not in updated:
                updated[key] = getattr(other, key)
        # Update model_extra
        if self.model_extra is not None:
            for key, value in self.model_extra.items():
                updated[key] = cast(PerFunctionPlotConfig, value)
                if other.model_extra is not None and key in other.model_extra:
                    updated[key] = updated[key].merge(
                        cast(PerFunctionPlotConfig, other.model_extra[key])
                    )
        if other.model_extra is not None:
            for key, value in other.model_extra.items():
                if key not in updated:
                    updated[key] = value

        return PlotConfig.model_validate(updated)


def create_parameter_docstring_mapping(func: Callable[..., Any]) -> Mapping[str, str]:
    """Create a mapping of parameter names and they docstrings.

    Parameters
    ----------
    func : Callable[..., Any]
        Function to create the parameter docstring mapping for.

    Returns
    -------
    Mapping[str, str]
    """
    param_docstring_mapping = {}
    for param in parse_docstring(func.__doc__ if func.__doc__ is not None else "").params:
        if param.description is not None:
            param_docstring_mapping[param.arg_name] = " ".join(param.description.splitlines())
    return param_docstring_mapping


def extract_default_kwargs(
    func: Callable[..., Any], exclude_kwargs: tuple[str, ...]
) -> DefaultKwargs:
    """Extract the default kwargs of ``func`` from its signature.

    Parameters
    ----------
    func : Callable[..., Any]
        Function to extract the default args from.
    exclude_kwargs : tuple[str, ...]
        Names of keyword arguments that should be excluded.

    Returns
    -------
    DefaultKwargs

    See Also
    --------
    use_plot_config
    """
    sig = signature(func)
    param_docstring_mapping = create_parameter_docstring_mapping(func)
    return {
        k: {
            "default": v.default,
            "annotation": v.annotation if v.annotation is not Parameter.empty else "object",
            "docstring": param_docstring_mapping.get(k, None),
        }
        for k, v in sig.parameters.items()
        if k not in exclude_kwargs
        and v.default is not Parameter.empty
        and v.kind is not Parameter.POSITIONAL_ONLY
    }


def find_not_user_provided_kwargs(
    default_kwargs: DefaultKwargs, arg_names: Iterable[str], kwargs: Mapping[str, Any]
) -> set[str]:
    """Find which kwargs of a function were not provided by the user.

    Those kwargs can be overridden by config value.

    Parameters
    ----------
    default_kwargs : DefaultKwargs
        Default keyword arguments to the function.
    arg_names : Iterable[str]
        Names of the positional arguments passed when calling the function.
    kwargs : Mapping[str, Any]
        Kwargs passed when calling the function.

    Returns
    -------
    set[str]

    See Also
    --------
    extract_default_kwargs
    """
    return {k for k in default_kwargs if k not in kwargs and k not in arg_names}


def find_axes(
    values: Iterable[Any],
) -> Generator[Axes, None, None]:
    """Iterate over values and yield the values that are ``Axes``.

    Parameters
    ----------
    values : Iterable[Any]
        Values to look for an ``Axes`` values in.

    Yields
    ------
    Axes
    """
    for value in values:
        if isinstance(value, str):
            continue
        elif isinstance(value, Axes):
            yield value
        elif isinstance(value, np.ndarray):
            yield from find_axes(value.flatten())
        elif isinstance(value, Iterable):
            yield from find_axes(value)


def use_plot_config(  # noqa: DOC201, DOC203
    exclude_from_config: tuple[str, ...] = (),
) -> Callable[[Callable[Param, RetType]], Callable[Param, RetType]]:
    """Decorate plot functions to register it and enables auto use of config.

    Parameters
    ----------
    exclude_from_config : tuple[str, ...]
        Names of keyword argument with default for which the type can not be represent in the
        config. Defaults to ()
    """

    def outer_wrapper(func: Callable[Param, RetType]) -> Callable[Param, RetType]:  # noqa: DOC
        """Outer wrapper to allow for ``ignore_kwargs`` to be passed."""
        default_kwargs = extract_default_kwargs(func, exclude_from_config)
        __PlotFunctionRegistry[func.__name__] = default_kwargs

        @wraps(func)
        def wrapper(*args: Param.args, **kwargs: Param.kwargs) -> RetType:  # noqa: DOC
            """Wrap function and apply config."""
            from pyglotaran_extras import CONFIG

            CONFIG.reload()

            arg_names = func.__code__.co_varnames[: len(args)]
            not_user_provided_kwargs = find_not_user_provided_kwargs(
                default_kwargs, arg_names, kwargs
            )
            function_config = CONFIG.plotting.get_function_config(func.__name__)
            override_kwargs = function_config.find_override_kwargs(not_user_provided_kwargs)
            updated_kwargs = kwargs | override_kwargs
            arg_axes = find_axes(getcallargs(func, *args, **updated_kwargs).values())
            return_values = func(*args, **updated_kwargs)
            function_config.update_axes_labels(arg_axes)

            if isinstance(return_values, Iterable):
                return_axes = find_axes(return_values)
                function_config.update_axes_labels(return_axes)

            return return_values

        return wrapper

    return outer_wrapper


@contextmanager
def plot_config_context(plot_config: PerFunctionPlotConfig) -> Generator[Config, None, None]:
    """Context manager to override parts of the resolved functions ``PlotConfig``.

    Parameters
    ----------
    plot_config : PerFunctionPlotConfig
        Function plot config override to update plot config for functions run inside of context.

    Yields
    ------
    Config
    """
    from pyglotaran_extras import CONFIG

    setattr(
        CONFIG.plotting,
        "__context_config",
        PerFunctionPlotConfig.model_validate(plot_config),
    )
    yield CONFIG
    delattr(CONFIG.plotting, "__context_config")



"""Inspection utility module."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Generator
    from collections.abc import Iterable


def wrap_in_details_tag(
    details_content: str,
    *,
    summary_content: str | None = None,
    summary_heading_level: int | None = None,
    is_open: bool = False,
) -> str:
    """Wrap ``details_content`` in a html details tag and add summary if ``summary_content`` set.

    Parameters
    ----------
    details_content : str
        Markdown string that should be displayed when the details are expanded.
    summary_content : str | None
        Summary test that should be displayed. Defaults to None so the summary is ``Details``.
    summary_heading_level : int | None
        Level of the heading wrapping the ``summary`` if it is not None. Defaults to None.
    is_open : bool
        Whether or not the details tag should be initially opened. Defaults to False.

    Returns
    -------
    str
    """
    out_str = f'\n<details {"open" if is_open else ""}>\n'
    if summary_content is not None:
        out_str += "<summary>\n"
        if summary_heading_level is None:
            out_str += f"{summary_content}\n"
        else:
            # Ref.:
            # https://css-tricks.com/two-issues-styling-the-details-element-and-how-to-solve-them/
            out_str += f'<h{summary_heading_level} style="display:inline;">\n'
            out_str += f"{summary_content}\n"
            out_str += f"</h{summary_heading_level}>\n"

        out_str += "</summary>\n"

    out_str += f"\n{details_content}\n<br>\n</details>"
    return out_str


def pretty_format_numerical(value: float | int, decimal_places: int = 1) -> str:
    """Format value with with at most ``decimal_places`` decimal places.

    Used to format values like the t-value.

    TODO : remove after raise pyglotaran dependency to 0.7.0
    Forward port of https://github.com/glotaran/pyglotaran/pull/1192

    Parameters
    ----------
    value : float | int
        Numerical value to format.
    decimal_places : int
        Decimal places to display. Defaults to 1.

    Returns
    -------
    str
        Pretty formatted version of the value.
    """
    # Bool returned by numpy do not support the ``is`` comparison (not same singleton as in python)
    # Ref: https://stackoverflow.com/a/37744300/3990615
    if not np.isfinite(value):
        return str(value)
    if abs(value - int(value)) <= np.finfo(np.float64).eps:
        return str(int(value))
    abs_value = abs(value)
    if abs_value < 10 ** (-decimal_places):
        format_instruction = f".{decimal_places}e"
    elif abs_value < 10 ** (decimal_places):
        format_instruction = f".{decimal_places}f"
    else:
        format_instruction = ".0f"
    return f"{value:{format_instruction}}"


def pretty_format_numerical_iterable(
    input_values: Iterable[str | float], decimal_places: int | None = 3
) -> Generator[str | float, None, None]:
    """Pretty format numerical values in an iterable of numerical values or strings.

    Parameters
    ----------
    input_values : Iterable[str | float]
        Values that should be formatted.
    decimal_places : int | None
        Number of decimal places a value should have, if None the original value will be used.
        Defaults to 3.

    See Also
    --------
    pretty_format_numerical

    Yields
    ------
    str | float
        Formatted string or initial value if ``decimal_places`` is None.
    """
    for val in input_values:
        if decimal_places is None or isinstance(val, str):
            yield val
        else:
            yield pretty_format_numerical(val, decimal_places=decimal_places)



"""Module containing data plotting functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

import matplotlib.pyplot as plt
from glotaran.io.prepare_dataset import add_svd_to_dataset
from matplotlib.axis import Axis

from pyglotaran_extras.config.plot_config import use_plot_config
from pyglotaran_extras.io.load_data import load_data
from pyglotaran_extras.plotting.plot_svd import plot_lsv_data
from pyglotaran_extras.plotting.plot_svd import plot_rsv_data
from pyglotaran_extras.plotting.plot_svd import plot_sv_data
from pyglotaran_extras.plotting.style import PlotStyle
from pyglotaran_extras.plotting.utils import MinorSymLogLocator
from pyglotaran_extras.plotting.utils import not_single_element_dims
from pyglotaran_extras.plotting.utils import shift_time_axis_by_irf_location

__all__ = ["plot_data_overview"]

if TYPE_CHECKING:
    from collections.abc import Hashable

    import xarray as xr
    from cycler import Cycler
    from glotaran.project.result import Result
    from matplotlib.figure import Figure
    from matplotlib.pyplot import Axes

    from pyglotaran_extras.types import DatasetConvertible


@use_plot_config(exclude_from_config=("svd_cycler",))
def plot_data_overview(
    dataset: DatasetConvertible | Result,
    title: str = "Data overview",
    linlog: bool = False,
    linthresh: float = 1,
    figsize: tuple[float, float] = (15, 10),
    nr_of_data_svd_vectors: int = 4,
    show_data_svd_legend: bool = True,
    irf_location: float | None = None,
    cmap: str = "PuRd",
    vmin: float | None = None,
    vmax: float | None = None,
    svd_cycler: Cycler | None = PlotStyle().cycler,
    use_svd_number: bool = False,
) -> tuple[Figure, Axes] | tuple[Figure, Axis]:
    """Plot data as filled contour plot and SVD components.

    Parameters
    ----------
    dataset : DatasetConvertible | Result
        Dataset containing data and SVD of the data.
    title : str
        Title to add to the figure. Defaults to "Data overview".
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    figsize : tuple[float, float]
        Size of the figure (N, M) in inches. Defaults to (15, 10).
    nr_of_data_svd_vectors : int
        Number of data SVD vector to plot. Defaults to 4.
    show_data_svd_legend : bool
        Whether or not to show the data SVD legend. Defaults to True.
    irf_location : float | None
        Location of the ``irf`` by which the time axis will get shifted. If it is None the time
        axis will not be shifted. Defaults to None.
    cmap : str
        Colormap to use for the filled contour plot. Defaults to "PuRd" which is most suitable
        for emission data (previous default was "viridis").
    vmin : float | None
        Lower value to anchor the colormap. Defaults to None meaning it inferred from the data.
    vmax : float | None
        Lower value to anchor the colormap. Defaults to None meaning it inferred from the data.
    svd_cycler : Cycler | None
        Plot style cycler to use for SVD plots. Defaults to ``PlotStyle().cycler``.
    use_svd_number : bool
        Whether to use singular value number (starts at 1) instead of singular value index
        (starts at 0) for labeling in plot. Defaults to False.

    Returns
    -------
    tuple[Figure, Axes] | tuple[Figure, Axis]
        Figure and axes which can then be refined by the user.
    """
    dataset = load_data(dataset, _stacklevel=3)
    data = shift_time_axis_by_irf_location(dataset.data, irf_location, _internal_call=True)

    if len(not_single_element_dims(data)) == 1:
        return _plot_single_trace(
            data,
            not_single_element_dims(data)[0],
            title="Single trace data",
            linlog=linlog,
            linthresh=linthresh,
            figsize=figsize,
        )

    fig = plt.figure(figsize=figsize)
    data_ax = cast(Axis, plt.subplot2grid((4, 3), (0, 0), colspan=3, rowspan=3, fig=fig))
    fig.subplots_adjust(hspace=0.5, wspace=0.25)
    lsv_ax = cast(Axis, plt.subplot2grid((4, 3), (3, 0), fig=fig))
    sv_ax = cast(Axis, plt.subplot2grid((4, 3), (3, 1), fig=fig))
    rsv_ax = cast(Axis, plt.subplot2grid((4, 3), (3, 2), fig=fig))

    if len(data.time) > 1:
        data.plot(x="time", ax=data_ax, center=False, cmap=cmap, vmin=vmin, vmax=vmax)
    else:
        data.plot(ax=data_ax)

    add_svd_to_dataset(dataset=dataset, name="data")
    plot_lsv_data(
        dataset,
        lsv_ax,
        indices=range(nr_of_data_svd_vectors),
        show_legend=False,
        linlog=linlog,
        linthresh=linthresh,
        irf_location=irf_location,
        cycler=svd_cycler,
        use_svd_number=use_svd_number,
    )
    plot_sv_data(dataset, sv_ax, use_svd_number=use_svd_number)
    plot_rsv_data(
        dataset,
        rsv_ax,
        indices=range(nr_of_data_svd_vectors),
        show_legend=False,
        cycler=svd_cycler,
        use_svd_number=use_svd_number,
    )
    if show_data_svd_legend is True:
        rsv_ax.legend(
            title="singular value number" if use_svd_number else "singular_value_index",
            loc="lower right",
            bbox_to_anchor=(1.13, 1),
        )
    fig.suptitle(title, fontsize=16)

    if linlog:
        data_ax.set_xscale("symlog", linthresh=linthresh)
        data_ax.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))
    return fig, (data_ax, lsv_ax, sv_ax, rsv_ax)


def _plot_single_trace(
    data_array: xr.DataArray,
    x_dim: Hashable,
    *,
    title: str = "Single trace data",
    linlog: bool = False,
    linthresh: float = 1,
    figsize: tuple[float, float] = (15, 10),
) -> tuple[Figure, Axis]:
    """Plot single trace data in case ``plot_data_overview`` gets passed ingle trace data.

    Parameters
    ----------
    data_array : xr.DataArray
        DataArray containing only data  of a single trace.
    x_dim : Hashable
        Name of the x dimension.
    title : str
        Title to add to the figure. Defaults to "Data overview".
    linlog : bool
        Whether to use 'symlog' scale or not. Defaults to False.
    linthresh : float
        A single float which defines the range (-x, x), within which the plot is linear.
        This avoids having the plot go to infinity around zero. Defaults to 1.
    figsize : tuple[float, float]
        Size of the figure (N, M) in inches. Defaults to (15, 10).

    Returns
    -------
    tuple[Figure, Axis]
        Figure and axis which can then be refined by the user.
    """
    fig, ax = plt.subplots(1, 1, figsize=figsize)
    data_array.plot.line(x=x_dim, ax=ax)
    fig.suptitle(title, fontsize=16)

    if linlog:
        ax.set_xscale("symlog", linthresh=linthresh)
        ax.xaxis.set_minor_locator(MinorSymLogLocator(linthresh))
    return fig, ax



"""Module containing predefined plot styles.

For reference see:
https://glotaran.github.io/legacy/plot_styles
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

import matplotlib.colors as colors
import matplotlib.pyplot as plt
from cycler import cycler

if TYPE_CHECKING:
    from cycler import Cycler


class ColorCode(str, Enum):
    """Color definitions from legacy glotaran.

    See: https://glotaran.github.io/legacy/plot_styles
    """

    # Name	#Hex
    black = "#000000"
    red = "#ff0000"
    blue = "#0000ff"
    green = "#00ff00"
    magenta = "#ff00ff"
    cyan = "#00ffff"
    yellow = "#ffff00"
    green4 = "#008b00"
    orange = "#ff8c00"
    brown = "#964b00"
    grey = "#808080"
    violet = "#9400d3"
    turquoise = "#40e0d0"
    maroon = "#800000"
    indigo = "#4b0082"

    @staticmethod
    def hex_to_rgb(hex_string: str) -> tuple[int, ...]:
        """Convert hex code to rgb or rgba tuple.

        Parameters
        ----------
        hex_string : str
            Hey code representation of a color.

        Returns
        -------
        tuple[int, ...]
            rgb or rgba tuple representing the same color as ``hex_string``..
        """
        rgb = colors.hex2color(hex_string)
        return tuple(int(255 * x) for x in rgb)

    @staticmethod
    def rgb_to_hex(rgb_tuple: tuple[float, ...]) -> str:
        """Convert rgb value tuple to hex code.

        Parameters
        ----------
        rgb_tuple : tuple[float, ...]
            Tuple rgb or rgba values

        Returns
        -------
        str
            Hex code representing the same color as ``rgb_tuple``.
        """
        return colors.rgb2hex([1.0 * x / 255 for x in rgb_tuple])


class LineStyle(str, Enum):
    """Subset of line styles supported by matplotlib."""

    solid = "-"
    dashed = "--"
    dotted = ".."
    dashdot = "-."


class DataColorCode(str, Enum):
    """Colors used to plot data and fits.

    Pairs of visually similar looking colors whereby the first (lighter) color is used to plot the
    data, and the second (darker) color is used to represent the fitted trace (that goes 'through'
    the data).
    """

    # Name	#Hex
    grey = "#808080"
    black = "#000000"
    orange = "#ff8c00"
    red = "#ff0000"
    cyan = "#00ffff"
    blue = "#0000ff"
    green = "#00ff00"
    green4 = "#008b00"
    magenta = "#ff00ff"
    indigo = "#4b0082"
    brown = "#964b00"
    maroon = "#800000"
    yellow = "#ffff00"
    # Adding orange a second time needs to be added manually to the list in PlotStyle
    # since Enum doesn't allow duplicates


class DataLineStyle(str, Enum):
    """Data plots can alternate between solid lines for data and dashed lines for fits.

    This is mostly useful for data with very low noise (e.g. simulated data), since data and fit
    often overlap.
    """

    solid = "-"
    dashed = "--"


class PlotStyle:
    """Wrapper class to hold predefined Plot styles."""

    def __init__(self) -> None:
        """Initialize Stiles from Enums."""
        self._line_style = list(LineStyle)
        self._color_codes = list(ColorCode)
        # Since Enum only supports unique values we need to manually add the last one
        self._data_color_codes = [*list(DataColorCode), DataColorCode.orange]
        # Extend linecycler to same size as colorcycler
        self._data_line_style = list(DataLineStyle) * (len(self._data_color_codes) // 2)
        self.SMALL_SIZE = 12
        self.MEDIUM_SIZE = 14
        self.BIGGER_SIZE = 18

    def set_default_fontsize(self) -> None:
        """Set global plot settings for matplotlib."""
        plt.rc("font", size=self.SMALL_SIZE)  # controls default text sizes
        plt.rc("axes", titlesize=self.SMALL_SIZE)  # fontsize of the axes title
        plt.rc("axes", labelsize=self.MEDIUM_SIZE)  # fontsize of the x and y labels
        plt.rc("xtick", labelsize=self.SMALL_SIZE)  # fontsize of the tick labels
        plt.rc("ytick", labelsize=self.SMALL_SIZE)  # fontsize of the tick labels
        plt.rc("legend", fontsize=self.SMALL_SIZE)  # legend fontsize
        plt.rc("figure", titlesize=self.BIGGER_SIZE)  # fontsize of the figure title

    def set_default_colors(self) -> None:
        """Set default color cycler for matplotlib."""
        plt.rc("axes", prop_cycle=self.cycler)

    @property
    def cycler(self) -> Cycler:
        """Cycle for default legacy like plots using color defined in :class:`ColorCode`.

        Returns
        -------
        Cycler
            Plot color cycler for general plots.
        """
        return cycler("color", self._color_codes)

    @property
    def data_cycler_solid_dashed(self) -> Cycler:
        """Cycle that alternates between solid and dashes and uses :class:`DataColorCode`.

        This is useful when plotting simulation or very low noise data, where the
        fit practically falls on top and hides the data lines.

        Returns
        -------
        Cycler
            Plot color cycler for low noise data+fit plots.
        """
        return cycler(color=self._data_color_codes, linestyle=self._data_line_style)

    @property
    def data_cycler_solid(self) -> Cycler:
        """Color cycler using :class:`DataColorCode`.

        Returns
        -------
        Cycler
            Plot color cycler for data+fit plots.
        """
        return cycler(color=self._data_color_codes)

    @property
    def line_cycler(self) -> Cycler:
        """Cycle iterating over line styles defined in :class:`LineStyle`.

        Returns
        -------
        Cycler
            Plot color cycler for multi line style plots.
        """
        return cycler("linestyle", self._line_style)



"""Io utility module."""

from __future__ import annotations

from collections.abc import Mapping
from collections.abc import Sequence
from pathlib import Path

import xarray as xr
from glotaran.project.result import Result

from pyglotaran_extras.io.load_data import load_data
from pyglotaran_extras.types import ResultLike


def result_dataset_mapping(result: ResultLike) -> Mapping[str, xr.Dataset]:
    """Convert a ``ResultLike`` object to a per dataset mapping of result like data.

    Parameters
    ----------
    result : ResultLike
        Data structure which can be converted to a mapping.

    Returns
    -------
    Mapping[str, xr.Dataset]
        Per dataset mapping of result like data.

    Raises
    ------
    TypeError
        If any value of a ``result`` isn't of :class:`DatasetConvertible`.
    TypeError
        If ``result`` isn't a :class:`ResultLike` object.
    """
    result_mapping = {}
    if isinstance(result, Result):
        return result.data
    if isinstance(result, xr.Dataset | xr.DataArray | Path | str):
        return {"dataset": load_data(result)}
    if isinstance(result, Sequence):
        for index, value in enumerate(result):
            result_mapping[f"dataset{index}"] = load_data(value)
        return result_mapping
    if isinstance(result, Mapping):
        for key, value in result.items():
            result_mapping[key] = load_data(value)
        return result_mapping
    msg = f"Result needs to be of type {ResultLike!r}, but was {result!r}."
    raise TypeError(msg)
